/*
 * EmulOS 2
 * PowerPC 750 CPU
 *
 * The PowerPC is a RISC CPU with __ GP registers
 * 
 * MMU:
 * - Can user Page-Level translation (4-KByte)
 * - or Block level translation (128-KByte to 256-MByte)
 */
#ifndef _COMMON_H_
#define _COMMON_H_

// === TYPES ===
typedef struct sMSR	tMSR;
typedef struct sSPRDef	tSPRDef;
typedef Uint32	tGPR;
typedef double	tFPR;
typedef struct sCPU	tCPU;

struct sMSR
{
	// TODO: Check the actual order
	unsigned	IR: 1;
	unsigned	DR: 1;
};

struct sSPRDef
{
	BOOL	UserAccessable;
	char	*Name;
}

struct sCPU
{
	Uint32	CR;	// Condition Register
	
	tFPR	FPRs[32];	// 64-bit registers
	Uint32	FPSCR;	// Floating Pointer Status and Control Register
	
	tGPR	GPRs[32];	// General Purpose Registers
	
	Uint32	SPRs[1024];	// Special Purpose Registers
	
	/*
	Uint32	LR;	// Link Register (SPR-8)
	Uint32	CTR;	// Counter (SPR-9)
	Uint32	DAR;	// Data Address Register(SPR-19)
	Uint32	DSISR;	// Data Error Code (SPR-18)
	Uint32	XER;	// Integer Exception Register (SPR-1)
	
	Uint32	SR[16];	// Segment Registers
	Uint32	BATs[16];	// Block Address Translation Registers (SPR-528 to SPR-535)
	
	Uint32	DABR;	// Data-Address Breakpoint Register (SPR-1013)
	Uint32	DEC;	// Decremeter - Rudimentary Timer (SPR-22)
	Uint32	EAR;	// External Access Register (SPR-282)
	
	Uint32	PVR;	// Processor Version Register - Read Only (SPR-?)
	
	Uint32	SDR1;	// Page Table Format Selector (SPR-?)
	Uint32	SRR0, SRR1;	// Save/Restore Registers (IP and MSR) (SPR-? & SPR-?)
	
	Uint32	SPRG[4];	// Free for use by the OS (SPR-? to SPR-?)
	
	Uint32	TB[2];	// Time Base (User RO, OS RW) (SPR-? & SPR-?)
	*/
	
	
	tMSR	MSR;	// Machine State Register
};

#endif
