/* EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * exceptions.c
 * - Exception Handling and Dispatching
 */
#include <emulos.h>
#include "common.h"

// === CODE ===
int Exception_TripleFault( tCPU *Regs )
{
	Regs->AtomicAccess = 0;
	return -1;
}

int Exception_DoubleFault( tCPU *Regs )
{
	Regs->AtomicAccess = 0;
	if( Regs->NestedExceptions == 2 )
		return Exception_TripleFault(Regs);
	
	Regs->NestedExceptions ++;
	return -1;
}

/**
 */
int Exception_GPF( tCPU *Regs, Uint16 ErrCode )
{
	Regs->AtomicAccess = 0;
	if( Regs->NestedExceptions == 1 )
		return Exception_DoubleFault(Regs);
	
	Regs->NestedExceptions ++;
	return -1;
}

/**
 */
int	Exception_PageFault(tCPU *Regs, Uint64 LinearAddress, Uint16 ErrorCode)
{
	Regs->AtomicAccess = 0;
	if( Regs->NestedExceptions == 1 )
		return Exception_DoubleFault(Regs);
	
	// Hmmm... I don't think this should happen
	if( Regs->NestedExceptions == 2 )
		return Exception_TripleFault(Regs);
	
	return 1;
}

/**
 * \fn int Exception_UndefinedOpcode( tCPU *Regs )
 */
int Exception_UndefinedOpcode( tCPU *Regs )
{
	Regs->AtomicAccess = 0;
	if( Regs->NestedExceptions == 1 )
		return Exception_DoubleFault(Regs);
	
	// Hmmm... I don't think this should happen
	if( Regs->NestedExceptions == 2 )
		return Exception_TripleFault(Regs);
	
	Regs->NestedExceptions ++;
	
	PushInterrupt(Regs);
	if(Regs->Mode == MODE_RM)
	{
		// TODO: Check what address is used on a #UD in real mode
		Regs->CS.Sel = 0xF000;
		Regs->CS.Base = 0xF0000;
		Regs->RIP = 0xFFF0;
	}
	
	if(Regs->Mode == MODE_PM || Regs->Mode == MODE_VM8086)
	{
		// TODO: Check TSS and load SS/xSP from it
		// TODO: Load Exception Vector
	}
	
	// Long Mode - Check what happens
	
	return 1;
}

/**
 */
void PushInterrupt(tCPU *Regs)
{
	if(Regs->Mode == MODE_RM)
	{
		WriteSegment16( Regs, &Regs->SS, Regs->GPR[REG_SP].X, GetFlags(Regs));
		Regs->GPR[REG_SP].X += 2;
		WriteSegment16( Regs, &Regs->SS, Regs->GPR[REG_SP].X, Regs->CS.Sel );
		Regs->GPR[REG_SP].X += 2;
		WriteSegment16( Regs, &Regs->SS, Regs->GPR[REG_SP].X, Regs->RIP );
		Regs->GPR[REG_SP].X += 2;
		return ;
	}
	
	if(Regs->Mode == MODE_PM || Regs->Mode == MODE_VM8086)
	{
		// TODO: Check TSS and load SS/xSP from it
		WriteSegment32( Regs, &Regs->SS, Regs->GPR[REG_SP].X, GetFlags(Regs));
		Regs->GPR[REG_SP].X += 4;
		WriteSegment32( Regs, &Regs->SS, Regs->GPR[REG_SP].X, Regs->CS.Sel );
		Regs->GPR[REG_SP].X += 4;
		WriteSegment32( Regs, &Regs->SS, Regs->GPR[REG_SP].X, Regs->RIP );
		Regs->GPR[REG_SP].X += 4;
	}
}
