/* EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * main.c
 * - Driver Core
 */
#include <emulos.h>
#include "common.h"

#define	MAX_VCPUS	1

// === IMPORTS ===
extern int	OpcodeDecode(tCPU *Regs);

// === PROTOTYPES ===
tCPU_Definition	*DriverInit(int NCPUs);
 int	DriverUpdate(void *Regs);
void	*DumpRegisters(void *CPU);

// === GLOBALS ===
void	*gCPUPtrs[MAX_VCPUS];
char	*gRegNames[10] = {"RAX","RCX","RDX","RBX","RSP","RBP","RSI","RDI","RIP","RFLAGS"};
Uint64	gRegDump[10];
tCPU	gCPUs[MAX_VCPUS];
tCPU_Definition	gCPU_Def = {
	"X86-64",
	DriverUpdate, NULL, DumpRegisters,
	64,
	0, gCPUPtrs,
	10,	gRegNames
};

// === CODE ===
tCPU_Definition *DriverInit(int NCPUs)
{
	 int	i;
	
	if( NCPUs > MAX_VCPUS )
		NCPUs = MAX_VCPUS;
	
	gCPU_Def.NumCPUs = NCPUs;
	
	for( i = 0; i < NCPUs; i++ )
	{
		gCPUs[i].Num = i;
		gCPUs[i].Mode = MODE_OFF;
		
		gCPUs[i].CS.Name = "CS";
		gCPUs[i].CS.Limit = 0xFFFF;
		gCPUs[i].CS.Flags.Code = TRUE;
		
		gCPUs[i].SS.Name = "SS";
		gCPUs[i].SS.Limit = 0xFFFF;
		gCPUs[i].DS.Name = "DS";
		gCPUs[i].DS.Limit = 0xFFFF;
		gCPUs[i].ES.Name = "ES";
		gCPUs[i].ES.Limit = 0xFFFF;
		gCPUs[i].FS.Name = "FS";
		gCPUs[i].FS.Limit = 0xFFFF;
		gCPUs[i].GS.Name = "GS";
		gCPUs[i].GS.Limit = 0xFFFF;
		
		gCPU_Def.CPUs[i] = &gCPUs[i];
	}
	
	gCPUs[0].Timestamp = 0;
	gCPUs[0].Debug.Trace = 1;
	
	gCPUs[0].Mode = MODE_RM;
	gCPUs[0].RIP = 0xFFF0;
	
	gCPUs[0].CS.Sel = 0xF000;
	gCPUs[0].CS.Base = 0xF0000;
	
	return &gCPU_Def;
}

int DriverUpdate(void *CPU)
{
	tCPU	*Regs = CPU;
	if( Regs->StateFlags.Halted )	return 0;
	if( Regs->Mode == MODE_OFF )	return 0;
	Emu_DumpRegs(CPU);
	return OpcodeDecode(Regs);
}

void *DumpRegisters(void *CPU)
{
	tCPU	*Regs = CPU;
	
	gRegDump[0] = Regs->GPR[0].R;
	gRegDump[1] = Regs->GPR[1].R;
	gRegDump[2] = Regs->GPR[2].R;
	gRegDump[3] = Regs->GPR[3].R;
	gRegDump[4] = Regs->GPR[4].R;
	gRegDump[5] = Regs->GPR[5].R;
	gRegDump[6] = Regs->GPR[6].R;
	gRegDump[7] = Regs->GPR[7].R;
	gRegDump[8] = Regs->RIP;
	gRegDump[9] = GetFlags(Regs);
	
	return gRegDump;
}

/**
 * \fn Uint32 GetFlags(tCPU *Regs)
 * \brief Converts the RFlags structure into a DoubleWord
 */
Uint32 GetFlags(tCPU *Regs)
{
	Uint16	ret = 0x0002;
	ret |= Regs->RFlags.C ? 0x0001 : 0;
	ret |= Regs->RFlags.P ? 0x0004 : 0;
	ret |= Regs->RFlags.A ? 0x0010 : 0;
	ret |= Regs->RFlags.Z ? 0x0040 : 0;
	ret |= Regs->RFlags.S ? 0x0080 : 0;
	ret |= Regs->RFlags.T ? 0x0100 : 0;
	ret |= Regs->RFlags.I ? 0x0200 : 0;
	ret |= Regs->RFlags.D ? 0x0400 : 0;
	ret |= Regs->RFlags.O ? 0x0800 : 0;
	ret |= (Regs->RFlags.IOPL & 3) << 12;
	ret |= Regs->RFlags.NT ? 0x4000 : 0;
	ret |= Regs->RFlags.R   ? 0x00010000 : 0;
	ret |= Regs->RFlags.VM  ? 0x00020000 : 0;
	ret |= Regs->RFlags.AC  ? 0x00040000 : 0;
	ret |= Regs->RFlags.VI  ? 0x00080000 : 0;
	ret |= Regs->RFlags.VIP ? 0x00100000 : 0;
	ret |= Regs->RFlags.ID  ? 0x00200000 : 0;
	return ret;
}

/**
 * \fn void SetFlags(tCPU *Regs, Uint64 Value, int Bits)
 */
void SetFlags(tCPU *Regs, Uint64 Value, int Bits)
{
	switch(Bits)
	{
	case 64:
	case 32:
		Regs->RFlags.R  = (Value & 0x00010000) ? TRUE : FALSE;
		Regs->RFlags.VM = (Value & 0x00020000) ? TRUE : FALSE;
		Regs->RFlags.AC = (Value & 0x00040000) ? TRUE : FALSE;
		Regs->RFlags.VI = (Value & 0x00080000) ? TRUE : FALSE;
		Regs->RFlags.VIP = (Value & 0x00100000) ? TRUE : FALSE;
		Regs->RFlags.ID = (Value & 0x00200000) ? TRUE : FALSE;
	case 16:
		Regs->RFlags.C  = (Value & 0x0001) ? TRUE : FALSE;
		Regs->RFlags.P  = (Value & 0x0004) ? TRUE : FALSE;
		Regs->RFlags.A  = (Value & 0x0010) ? TRUE : FALSE;
		Regs->RFlags.Z  = (Value & 0x0040) ? TRUE : FALSE;
		Regs->RFlags.S  = (Value & 0x0080) ? TRUE : FALSE;
		Regs->RFlags.T  = (Value & 0x0100) ? TRUE : FALSE;
		Regs->RFlags.I  = (Value & 0x0200) ? TRUE : FALSE;
		Regs->RFlags.D  = (Value & 0x0400) ? TRUE : FALSE;
		Regs->RFlags.O  = (Value & 0x0800) ? TRUE : FALSE;
		Regs->RFlags.IOPL = (Value >> 12) & 3;
		Regs->RFlags.NT = (Value & 0x4000) ? TRUE : FALSE;
		break;
	}
}

/**
 * \fn Uint32 GetCR0(tCPU *Regs)
 * \brief Packs the CR0 Register into a DWORD
 */
Uint32 GetCR0(tCPU *Regs)
{
	Uint32	ret = 0;
	ret |= Regs->CR0.PE ? 0x00000001 : 0;
	ret |= Regs->CR0.MP ? 0x00000002 : 0;
	ret |= Regs->CR0.EM ? 0x00000004 : 0;
	ret |= Regs->CR0.TS ? 0x00000008 : 0;
	ret |= Regs->CR0.ET ? 0x00000010 : 0;
	ret |= Regs->CR0.NE ? 0x00000020 : 0;
	ret |= Regs->CR0.WP ? 0x00010000 : 0;
	ret |= Regs->CR0.AM ? 0x00040000 : 0;
	ret |= Regs->CR0.NW ? 0x20000000 : 0;
	ret |= Regs->CR0.CD ? 0x40000000 : 0;
	ret |= Regs->CR0.PG ? 0x80000000 : 0;
	return ret;
}

/**
 * \fn void SetCR0(tCPU *Regs, Uint32 Value)
 * \brief Sets the CR0 structure from a DWord
 */
void SetCR0(tCPU *Regs, Uint32 Value)
{
	//Debug("SetCR0(..., Value=0x%8)\n");
	Regs->CR0.PE = (Value & 0x00000001) ? TRUE : FALSE;
	Regs->CR0.MP = (Value & 0x00000002) ? TRUE : FALSE;
	Regs->CR0.EM = (Value & 0x00000004) ? TRUE : FALSE;
	Regs->CR0.TS = (Value & 0x00000008) ? TRUE : FALSE;
	Regs->CR0.ET = (Value & 0x00000010) ? TRUE : FALSE;
	Regs->CR0.NE = (Value & 0x00000020) ? TRUE : FALSE;
	Regs->CR0.WP = (Value & 0x00010000) ? TRUE : FALSE;
	Regs->CR0.AM = (Value & 0x00040000) ? TRUE : FALSE;
	Regs->CR0.NW = (Value & 0x20000000) ? TRUE : FALSE;
	Regs->CR0.CD = (Value & 0x40000000) ? TRUE : FALSE;
	Regs->CR0.PG = (Value & 0x80000000) ? TRUE : FALSE;
}

/**
 * \fn Uint32 GetCR4(tCPU *Regs)
 */
Uint32 GetCR4(tCPU *Regs)
{
	Uint32	ret = 0;
	ret |= Regs->CR4.VME    ? 0x00000001 : 0;
	ret |= Regs->CR4.PVI    ? 0x00000002 : 0;
	ret |= Regs->CR4.TSD    ? 0x00000004 : 0;
	ret |= Regs->CR4.DE     ? 0x00000008 : 0;
	ret |= Regs->CR4.PSE    ? 0x00000010 : 0;
	ret |= Regs->CR4.PAE    ? 0x00000020 : 0;
	ret |= Regs->CR4.MCE    ? 0x00000040 : 0;
	ret |= Regs->CR4.PGE    ? 0x00000080 : 0;
	ret |= Regs->CR4.PCE    ? 0x00000100 : 0;
	ret |= Regs->CR4.OSFXSR ? 0x00000200 : 0;
	ret |= Regs->CR4.OSXMMEXCPT ? 0x00000400 : 0;
	// 0
	// 0
	ret |= Regs->CR4.VMXE   ? 0x00002000 : 0;
	ret |= Regs->CR4.SMXE   ? 0x00004000 : 0;
	ret |= Regs->CR4.OSXSAVE? 0x00008000 : 0;
	return ret;
}

/**
 * \fn void SetCR4(tCPU *Regs, Uint32 Value)
 * \brief Set CR4 from a DWord
 */
void SetCR4(tCPU *Regs, Uint32 Value)
{
	Regs->CR4.VME = (Value & 0x00000001) ? TRUE : FALSE;
	Regs->CR4.PVI = (Value & 0x00000002) ? TRUE : FALSE;
	Regs->CR4.TSD = (Value & 0x00000004) ? TRUE : FALSE;
	Regs->CR4.DE  = (Value & 0x00000008) ? TRUE : FALSE;
	Regs->CR4.PSE = (Value & 0x00000010) ? TRUE : FALSE;
	Regs->CR4.PAE = (Value & 0x00000020) ? TRUE : FALSE;
	Regs->CR4.MCE = (Value & 0x00000040) ? TRUE : FALSE;
	Regs->CR4.PGE = (Value & 0x00000080) ? TRUE : FALSE;
	Regs->CR4.PCE = (Value & 0x00000100) ? TRUE : FALSE;
	Regs->CR4.OSFXSR = (Value & 0x00000200) ? TRUE : FALSE;
	Regs->CR4.OSXMMEXCPT = (Value & 0x00000400) ? TRUE : FALSE;
	// 0
	// 0
	Regs->CR4.VMXE = (Value & 0x00002000) ? TRUE : FALSE;
	Regs->CR4.SMXE = (Value & 0x00004000) ? TRUE : FALSE;
	Regs->CR4.OSXSAVE = (Value & 0x00008000) ? TRUE : FALSE;
}
