/*EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * opcodes.h
 * - Opcode Defintions and Ancilary Byte Structures
 */

#ifndef _OPCODES_H_
#define _OPCODES_H_

typedef struct sModRM	tModRM;
typedef struct sSIB	tSIB;
typedef struct sREX	tREX;

enum {
	PREF_LOCK	= 0xF0,
	PREF_REPNZ	= 0xF2,
	PREF_REPZ	= 0xF3,
	
	PREF_CS	= 0x2E,	PREF_SS	= 0x36,
	PREF_DS	= 0x3E,	PREF_ES	= 0x26,
	PREF_FS	= 0x64,	PREF_GS	= 0x65,
	
	/*PREF_NOTTAKEN	= 0x2E,
	PREF_TAKEN	= 0x3E,*/
	
	PREF_OPRSIZE	= 0x66,
	PREF_ADDRSIZE	= 0x67,
	
	PREF_REX0	= 0x40,	PREF_REX1,
	PREF_REX2,	PREF_REX3,	PREF_REX4,	PREF_REX5,
	PREF_REX6,	PREF_REX7,	PREF_REX8,	PREF_REX9,
	PREF_REX10,	PREF_REX11,	PREF_REX12,	PREF_REX13,
	PREF_REX14,	PREF_REX15	= 0x4F,
	
	// --- Miscelaneous
	OP_HLT	= 0xF4,
	
	// --- Flow Control
	OP_CALL_NEAR = 0xE8,
	//OP_CALL_RM = 0xFF,	// /2
	OP_CALL_FAR = 0x9A,
	OP_JMP_FAR = 0xEA,
	OP_JMP_REL = 0xEB,
	OP_JMP_RELX = 0xE9,
	// --- Conditional Jumps
	OP_JCC_REL = 0x70,
	
	// --- Stack
	//OP_PUSH_RM = 0xFF,	// /6
	OP_PUSH_R = 0x50,	// To 0x57
	OP_PUSH_I = 0x6A,
	OP_PUSH_IX = 0x68,
	OP_PUSH_CS = 0x0E,	// #UD in 64-bit
	OP_PUSH_SS = 0x16,	// #UD in 64-bit
	OP_PUSH_DS = 0x1E,	// #UD in 64-bit
	OP_PUSH_ES = 0x06,	// #UD in 64-bit
	OP_PUSHA = 0x60,	// #UD in 64-bit
	OP_PUSHF = 0x9C,
	
	OP_POP_RM = 0x8F,	// /0
	OP_POP_R = 0x58,	// To 0x5F
	OP_POP_DS = 0x1F,	// NE in 64bit
	OP_POP_ES = 0x07,	// NE in 64bit
	OP_POP_SS = 0x17,	// NE in 64bit
	OP_POPA = 0x61,	// #UD in 64bit
	OP_POPF = 0x9D,
	
	// --- Mov
	OP_MOV_MR = 0x88,	OP_MOV_MRX = 0x89,	// mov r/m8, r8
	OP_MOV_RM = 0x8A,	OP_MOV_RMX = 0x8B,	// mov r8, r/m8
	OP_MOV_MS = 0x8C,	// mov r/m16, sreg
	OP_MOV_SM = 0x8E,	// mov sreg, r/m16
	OP_MOV_Am = 0xA0,	OP_MOV_AmX = 0xA1,	// mov (AL,AX,EAX,RAX), ds:moffs16
	OP_MOV_mA = 0xA2,	OP_MOV_mAX = 0xA3,	// mov ds:moffs16, (AL,AX,EAX,RAX)
	OP_MOV_RI = 0xB0,	// 8 opcode span - mov r8, imm8
	OP_MOV_RIX = 0xB8,	// 8 opcode span - mov r(16,32,64), imm(16,32,64)
	OP_MOV_MI = 0xC6,	// mov r/m8, imm8
	OP_MOV_MIX = 0xC7,	// mov r/m16, imm16
	
	//OP_MOV_MI = 0xC6	// /0
	//OP_MOV_MIX = 0xC7	// /0
	
	// --- Arithmatic
	OP_ARITH_MI = 0x80, OP_ARITH_MIX,
	OP_ARITH_MI8= 0x83,
	// - ADD
	OP_ADD_AI = 0x04, OP_ADD_AIX,
	OP_ADD_MR = 0x00, OP_ADD_MRX,
	OP_ADD_RM = 0x02, OP_ADD_RMX,
	// - OR
	OP_OR_AI  = 0x0C, OP_OR_AIX,
	OP_OR_MR  = 0x08, OP_OR_MRX,
	OP_OR_RM  = 0x0A, OP_OR_RMX,
	// - AND
	OP_AND_AI = 0x24, OP_AND_AIX,
	OP_AND_MR = 0x20, OP_AND_MRX,
	OP_AND_RM = 0x22, OP_AND_RMX,
	
	// - DEC
	OP_DEC_R = 0x48,	// 8 opcode span - dec r16
	
	// ===
	// 0x0F Extended Opcodes
	// ===
	OP_0F_PUSH_FS = 0xA0,
	OP_0F_PUSH_GS = 0xA8,
	OP_0F_POP_FS = 0xA1,
	OP_0F_POP_GS = 0xA9,
	
	OP_UNDEFINED = 0xFF
};

struct sModRM
{
	unsigned RM:	3;
	unsigned Reg:	3;
	unsigned Mod:	2;
};

struct sSIB
{
	unsigned Base:	3;
	unsigned Index:	3;
	unsigned Scale:	2;
};

struct sREX
{
	unsigned B:	1;	// Generic Extension (ModRM's R/M, SIB's base, Opcode's reg)
	unsigned X:	1;	// Extends SIB's Index Field
	unsigned R:	1;	// Extends ModRM's Reg Field
	unsigned W:	1;	// 0: Use CD.D, 1: Use 64-bit
	unsigned Valid:	4;	// 0100
};

#endif
