/* EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * voodoo.h
 * - Decoding helper macros
 */

#ifndef _VOODOO_H_
#define _VOODOO_H_

#define CASE8(base)	case(base):case((base)+1):case((base)+2):case((base)+3):\
	case((base)+4):case((base)+5):case((base)+6):case((base)+7)
#define CASE16(base)	CASE8(base):CASE8((base)+8)

#define USE_32BIT_OPERAND	(((Regs->Mode == MODE_PM || Regs->Mode == MODE_COMPAT) && !Regs->OvrOperand)\
	||((Regs->Mode == MODE_RM || Regs->Mode == MODE_VM8086) && Regs->OvrOperand)\
	||(Regs->Mode == MODE_LM && !Regs->REX.W))
#define USE_EXTENDTO64_OPERAND	(Regs->Mode == MODE_LM && Regs->REX.W)

#define GETMODRM()	ProcessModRM( Regs, &ipOffset, &mod, &reg )

// HACK: Requires the virtual address space to be less than 64 bits wide
#define IS_REG(_addr)	(((_addr)>>32)==0x80000000)
#define TO_REGADDR(_num)	(0x8000000000000000ll|(_num))
/**
 * GET_RM{64,32,16,8}
 * - Reads either a register or a memory location into `val{64,32,16,8}`
 * SET_RM{64,32,16,8}
 * - Writes `val{64,32,16,8}` back to register or memory
 * 
 * Uses:
 *  `addr` - Address/Register Index
 *  `Regs` - tCPU Structure
 *  `val...` - Destination/Source for data, depends on the size requested
 */
#define GET_RM64()	val64=(IS_REG(addr)?GetRegister64(Regs,addr&15):ReadSegment64(Regs,Regs->Seg,addr))
#define GET_RM32()	val32=(IS_REG(addr)?GetRegister32(Regs,addr&15):ReadSegment32(Regs,Regs->Seg,addr))
#define GET_RM16()	val16=(IS_REG(addr)?GetRegister16(Regs,addr&15):ReadSegment16(Regs,Regs->Seg,addr))
#define GET_RM8()	val8=(IS_REG(addr)?GetRegister8(Regs,addr&15):ReadSegment8(Regs,Regs->Seg,addr))
#define SET_RM64() do {\
	if(IS_REG(addr))	SetRegister64(Regs,addr&15,val64);\
	else	WriteSegment64(Regs,Regs->Seg,addr,val64);\
	}while(0)
#define SET_RM32() do {\
	if(IS_REG(addr))	SetRegister32(Regs,addr&15,val32);\
	else	WriteSegment32(Regs,Regs->Seg,addr,val32);\
	}while(0)
#define SET_RM16() do {\
	if(IS_REG(addr))	SetRegister16(Regs,addr&15,val16);\
	else	WriteSegment16(Regs,Regs->Seg,addr,val16);\
	}while(0)
#define SET_RM8() do {\
	if(IS_REG(addr))	SetRegister8(Regs,addr&15,val8);\
	else	WriteSegment8(Regs,Regs->Seg,addr,val8);\
	}while(0)

// --- ALU Functions ---

// #0: Add
#define DO_ADD(to, from, width)	do {\
	(to) += (from);\
	Regs->RFlags.Z = ((to) == 0);\
	Regs->RFlags.S = ((to) >> ((width)-1));\
	Regs->RFlags.C = Regs->RFlags.O = ((to) < (from));\
	SET_PF((to),(width));\
	}while(0)
// #1: Bitwise OR
#define DO_OR(to, from, width)	do {\
	(to) |= (from);\
	Regs->RFlags.Z = ((to) == 0);\
	Regs->RFlags.S = ((to) >> ((width)-1));\
	Regs->RFlags.C = Regs->RFlags.O = 0;\
	SET_PF((to),(width));\
	}while(0)
// #4: Bitwise AND
#define DO_AND(to, from, width)	do {\
	(to) &= (from);\
	Regs->RFlags.Z = ((to) == 0);\
	Regs->RFlags.S = ((to) >> ((width)-1));\
	Regs->RFlags.C = Regs->RFlags.O = 0;\
	SET_PF((to),(width));\
	}while(0)
//PF should be set ((((to)>>7)&1)^(((to)>>6)&1)^(((to)>>5)&1)^(((to)>>4)&1)^(((to)>>3)&1)^(((to)>>2)&1)^(((to)>>1)&1)^((to)&1))

#define PAIRITY8(v)	((((v)>>7)&1)^(((v)>>6)&1)^(((v)>>5)&1)^(((v)>>4)&1)^(((v)>>3)&1)^(((v)>>2)&1)^(((v)>>1)&1)^((v)&1))
#define PAIRITY32(v)	(PAIRITY8(v) ^ PAIRITY8((v)>>8) ^ PAIRITY8((v)>>16) ^ PAIRITY8((v)>>24))
#define SET_PF(v,w) do{\
	switch(w){\
	case 8:	Regs->RFlags.PF = PAIRITY8(v);	break;\
	case 16:	Regs->RFlags.PF = PAIRITY8(v) ^ PAIRITY8(v>>8);	break;\
	case 32:	Regs->RFlags.PF = PAIRITY32(v);	break;\
	case 64:	Regs->RFlags.PF = PAIRITY32(v) ^ PAIRITY32(v>>32);	break;\
	}\
	}while(0)

#endif
