/* EmulOS2 PC Driver
 * 
 * By thePowersGang
 */
#include <emulos.h>

// === PROTOTYPES ===
BOOL	RAM_Read(tEmuMem_Region *Region, Uint64 Offset, int Size, void *Data);
BOOL	RAM_Write(tEmuMem_Region *Region, Uint64 Offset, int Size, void *Data);

// === ROMs ===
extern const Uint8	_binary_bios_rom_start[];

// === GLOBALS ===
tMachine_Definition	gMachineDef = {
	"PC", 1
};
tEmuMem_Region	gMemory_LowRAM = {
	NULL, VRAM_START,
	0x00000,	0xA0000,
	RAM_Read,	RAM_Write
};
tEmuMem_Region	gMemory_BIOSROM = {
	NULL, (Uint)_binary_bios_rom_start,
	0xF0000,	0x10000,
	RAM_Read,	NULL
};

// === CODE ===
/**
 * \fn tMachine_Definition *DriverInit()
 * \brief Initialises the Machine driver
 * \return Number of CPUs required
 */
tMachine_Definition *DriverInit()
{
	// Allocate Low RAM
	MM_AllocateBlock(VRAM_START, 0xA0000);
	Memory_Register( &gMemory_LowRAM );
	Memory_Register( &gMemory_BIOSROM );
	
	return &gMachineDef;
}

/**
 */
BOOL RAM_Read(tEmuMem_Region *Region, Uint64 Offset, int Size, void *Data)
{
	void	*buf = (void*)(Region->Impl + (Uint32)Offset);
	//Debug("RAM_Read: (Region={Impl:%8},Offset=%X,Size=%i,Data=%8)\n",
	//	Region->Impl, Offset, Size, Data );
	switch(Size)
	{
	case 0:	break;
	case 1:	*(Uint8*)Data = *(Uint8*)buf;	break;
	case 2:	*(Uint16*)Data = *(Uint16*)buf;	break;
	case 4:	*(Uint32*)Data = *(Uint32*)buf;	break;
	case 8:	*(Uint64*)Data = *(Uint64*)buf;	break;
	default:	memcpy(Data, buf, Size);	break;
	}
	return TRUE;
}
/**
 */
BOOL RAM_Write(tEmuMem_Region *Region, Uint64 Offset, int Size, void *Data)
{
	void	*buf = (void*)(Region->Impl + (Uint32)Offset);
	switch(Size)
	{
	case 0:	break;
	case 1:	*(Uint8*)buf = *(Uint8*)Data;	break;
	case 2:	*(Uint16*)buf = *(Uint16*)Data;	break;
	case 4:	*(Uint32*)buf = *(Uint32*)Data;	break;
	case 8:	*(Uint64*)buf = *(Uint64*)Data;	break;
	default:	memcpy(buf, Data, Size);	break;
	}
	return TRUE;
}
