/*
 * EmulOS 2
 * emulos.h
 * - Common Header File
 */
#ifndef _EMULOS_H_
#define _EMULOS_H_

#define NULL	((void*)0)
#define TRUE	1
#define FALSE	0

// --- Host Architecture Depenent Defines ---
#define VRAM_START	0x100000	// 1MiB

// --- Common Types ---
typedef char	BOOL;
typedef unsigned int	size_t;

typedef unsigned int	Uint;
typedef signed int	Sint;

typedef unsigned char	Uint8;
typedef unsigned short	Uint16;
typedef unsigned long	Uint32;
typedef unsigned long long	Uint64;

typedef signed char	Sint8;
typedef signed short	Sint16;
typedef signed long	Sint32;
typedef signed long long	Sint64;

// --- Memory Access ---
typedef struct sEmuMem_Region	tEmuMem_Region;
typedef BOOL (*tEmuMem_ReadWrite)(tEmuMem_Region *Region, Uint64 Offset, int Size, void *Data);
struct sEmuMem_Region
{
	tEmuMem_Region	*Next;
	Uint	Impl;
	Uint64	Base;
	Uint64	Size;
	tEmuMem_ReadWrite	Read;
	tEmuMem_ReadWrite	Write;
};

// Real Memory
extern BOOL	MM_AllocateBlock(Uint Base, Uint Size);
// Emulated Memory
extern BOOL	Memory_Register(tEmuMem_Region *Region);
extern BOOL	Memory_Read(int Lock, Uint64 Address, int Size, void *Data);
extern BOOL	Memory_Write(int Lock, Uint64 Address, int Size, void *Data);
extern int	Memory_LockBus();
extern void	Memory_UnlockBus(int Lock);
// Utilities
extern void	Debug(char *Format, ...);
extern void	Emu_DumpRegs(void *CPU);
extern void	*memcpy(void *dest, const void *src, size_t size);

// --- Machine ---
typedef struct sMachine_Definition	tMachine_Definition;
struct sMachine_Definition
{
	char	*Name;
	 int	NumCPUs;
};

// --- CPU ---
typedef void* (*tCPU_GetRegs)(void *);
typedef int (*tCPU_Update)(void *);
typedef int (*tCPU_Deinit)();
typedef struct sCPU_Definition	tCPU_Definition;
struct sCPU_Definition
{
	char	*Name;

	tCPU_Update	Update;
	tCPU_Deinit	Cleanup;
	tCPU_GetRegs	GetRegs;
	
	 int	NumBits;
	 int	NumCPUs;
	void	**CPUs;
	
	 int	NumRegisters;
	char	**RegisterNames;
};

#endif
