; EmulOS 2
; By thePowersgang
; bin_elf.asm
; - ELF Binary Loader
[BITS 32]
%include "common.inc.asm"
%include "binary.inc.asm"
%include "bin_elf.inc.asm"

[extern MM_IsValid]
[extern MM_Allocate]
[extern strcmp]
[extern Binary_GetSymbol]

%macro RET_ZERO_CC	1
	j%1 %%rz
	jmp %%skip
%%rz:
	xchg bx,bx
	xor eax, eax
	jmp .ret
%%skip:
%endmacro

[section .text]
;; ELF_Load
;; \brief Load an ELF binary directly into memory
;; \note Does checking specific to EmulOS's memory layout
;; \param Address	Address of binary image in memory
[global ELF_Load]
ELF_Load:
	; 3 Vars, 2 Params
	FRAME_OPEN	4, 2, {esi, edi}
	
	; Get Address
	mov esi, ARG1
	
	; First, check for a program header
	mov edi, [esi+tELF_Header.PHdrOff]
	test edi, edi
	RET_ZERO_CC	z
	;jz .retZero
	add edi, esi

	; Get the base address and the limit
	xor ecx, ecx
	mov cx, WORD [esi+tELF_Header.phentcount]
	;xchg bx, bx
	test ecx, ecx
	RET_ZERO_CC	z
	;jz	.retZero
	
	
	mov DWORD VAR1, -1
	mov DWORD VAR2, 0
	
	;xchg bx, bx
	
.getrange.top:
	test ecx, ecx
	jz	.getrange.end
	
	cmp DWORD [edi+tELF_PhEnt.Type], 2	; PT_DYNAMIC
	jnz .getrange.notDyn
	mov eax, [edi+tELF_PhEnt.VAddr]
	mov VAR3, eax
	jmp	.getrange.next
.getrange.notDyn:
	
	; Check Type
	cmp DWORD [edi+tELF_PhEnt.Type], 1
	jnz	.getrange.next
	
	; Check for lowest address
	mov eax, [edi+tELF_PhEnt.VAddr]
	cmp eax, VAR1
	jae .getrange.notLowest
	mov VAR1, eax
.getrange.notLowest:
	; Check for highest address
	add eax, [edi+tELF_PhEnt.MemSize]
	cmp eax, VAR2
	jbe .getrange.notHighest
	mov VAR2, eax
.getrange.notHighest:

	; Next please!
.getrange.next:
	dec ecx
	jz .getrange.end
	xor eax, eax
	mov ax, WORD [esi+tELF_Header.phentsize]
	add edi, eax
	jmp .getrange.top
.getrange.end:

	; Ok, VAR1 is now the lowest addres
	; VAR2 is now the highest address

	; Second, check if the address space is sane
	mov eax, VAR1
	cmp eax, BIN_LOAD_LOWEST
	RET_ZERO_CC	b
	;jnz	.retZero
	mov eax, VAR2
	cmp eax, BIN_LOAD_HIGHEST
	RET_ZERO_CC	a
	sub eax, VAR1
	cmp eax, BIN_LOAD_QUANTUM
	;xchg bx, bx
	RET_ZERO_CC	a
	;ja	.retZero
	
	; Ok, it's sane, but is it free?
	mov eax, VAR1
.testLoop:
	mov [esp], eax
	call MM_IsValid
	test eax, eax
	RET_ZERO_CC	nz
	mov eax, [esp]
	add eax, 0x1000
	cmp eax, VAR2
	jl	.testLoop

	; Now, scan it, loading the segments into memory
	mov edi, [esi+tELF_Header.PHdrOff]
	add edi, esi
	xor ecx, ecx
	mov cx, WORD [esi+tELF_Header.phentcount]
	xor eax, eax
	
.load.top:
	test ecx, ecx
	jz	.load.end
	
	; Check Type
	cmp DWORD [edi+tELF_PhEnt.Type], 1
	jnz	.load.next
	
	mov VAR4, ecx	; Save ECX
	mov [esp], esi
	mov [esp+4], edi
	;DUMP_STACK	"ELF_INT_LoadSegment", "Base", "Segment"
	call ELF_INT_LoadSegment
	test eax, eax
	RET_ZERO_CC	nz
	;jnz	.retZero
	mov ecx, VAR4	; Save ECX
	
.load.next:
	dec ecx
	; EAX is zero here
	mov ax, WORD [esi+tELF_Header.phentsize]
	add edi, eax
	jmp .load.top
.load.end:
	
	
	; Now we're loaded, let's make sure we don't segfault when
	; ELF_GetSymbol is called
	mov esi, VAR1
	mov edi, VAR3
	test edi, edi
	jz .noDYNAMIC
	
.prescan.top:
	xor eax, eax
	mov ax, WORD [edi+tELF_Dyn.Type]
	cmp eax, .prescan.jumptable_size
	jae .prescan.next
	mov eax, [.prescan.jumptable+eax*4]
	jmp eax
[section .data]
.prescan.jumptable:
	dd	.prescan.end	; DT_NULL
	dd	.prescan.next	; DT_NEEDED
	dd	.prescan.next	; DT_PLTRELSZ
	dd	.prescan.next	; DT_PLTGOT
	dd	.prescan.hashTab	; DT_HASH
	dd	.prescan.strTab	; DT_STRTAB
	dd	.prescan.symTab	; DT_SYMTAB
.prescan.jumptable_len	equ	($-.prescan.jumptable)
.prescan.jumptable_size	equ	($-.prescan.jumptable)/4
[section .text]
.prescan.hashTab:
	mov eax, DWORD [edi+tELF_Dyn.Value]
	mov DWORD [esi+tELF_Header.dynhash], eax
	mov ecx, DWORD [eax+4]	; Get Symbol Count
	jmp .prescan.next
.prescan.strTab:
	mov edx, DWORD [edi+tELF_Dyn.Value]
	jmp .prescan.next
.prescan.symTab:
	mov eax, DWORD [edi+tELF_Dyn.Value]
	mov DWORD [esi+tELF_Header.dynsym], eax
	jmp .prescan.next
.prescan.next:
	add edi, tELF_Dyn_size
	jmp .prescan.top
.prescan.end:
	; EDX - String Table
	; ECX - Symbol Count
	; Ok, now we have the symbol table, string table and the hash table
	; Next, relocate the symbol table
	mov edi, DWORD [esi+tELF_Header.dynsym]
	
	; Sanity check please!
	test edi, edi
	jz .relocSyms.end
	test ecx, ecx
	jz .relocSyms.end
	
	; Relocate symbol names
.relocSyms.top:
	add DWORD [edi+tELF_Sym.Name], edx
.relocSyms.next:
	dec ecx
	jz .relocSyms.end
	add edi, tELF_Sym_size
	jmp .relocSyms.top
.relocSyms.end:

	; We're all ready to be used, let's return
	jmp .prepRet
	
.noDYNAMIC:
	mov DWORD [esi+tELF_Header.dynsym], 0
	mov DWORD [esi+tELF_Header.dynhash], 0

.prepRet:
	; Prepare for return
	mov eax, VAR1
	mov ecx, VAR3	; Location of .dynamic in the image
	mov [eax+tELF_Header.DYNAMIC], ecx
	jmp	.ret

.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; ELF_Relocate
;; \brief Relocates a loaded binary
;; \todo Move the symbol relocation and .dynsym and .hash finding to ELF_Load
[global ELF_Relocate]
ELF_Relocate:
	FRAME_OPEN	10, 4, {esi, edi}
	
	; VAR1	- String Table (UNUSED)
	mov DWORD VAR2, 0	; PLT Base
	mov DWORD VAR3, 0	; PLT Type
	mov DWORD VAR4, 0	; PLT Size
	mov DWORD VAR5, 0	; RELA Base
	mov DWORD VAR6, 0	; RELA Size
	mov DWORD VAR7, tELF_Rela_size	; RELA Entry Size
	mov DWORD VAR8, 0	; REL Base
	mov DWORD VAR9, 0	; REL Size
	mov DWORD VAR10, tELF_Rel_size	; REL Entry Size
	
	mov esi, DWORD ARG1
	mov edi, DWORD [esi+tELF_Header.DYNAMIC]
	test edi, edi
	jz .retZero

.prescan.top:
	xor eax, eax
	mov ax, WORD [edi+tELF_Dyn.Type]
	cmp eax, .prescan.jumptable_size
	jae .prescan.next
	mov eax, [.prescan.jumptable+eax*4]
	jmp eax

.prescan.jumptable:
	dd	.prescan.end	; DT_NULL
	dd	.prescan.next	; DT_NEEDED
	dd	.prescan.pltrelsz	; DT_PLTRELSZ
	dd	.prescan.next	; DT_PLTGOT
	dd	.prescan.next	; DT_HASH
	dd	.prescan.next	; DT_STRTAB
	dd	.prescan.next	; DT_SYMTAB
	dd	.prescan.rela	; DT_RELA
	dd	.prescan.relasz	; DT_RELASZ
	dd	.prescan.relaent	; DT_RELAENT
	dd	.prescan.next	; DT_STRSZ
	dd	.prescan.next	; DT_SYMENT
	dd	.prescan.next	; DT_INIT
	dd	.prescan.next	; DT_FINI
	dd	.prescan.next	; DT_SONAME
	dd	.prescan.next	; DT_RPATH
	dd	.prescan.next	; DT_SYMBOLIC
	dd	.prescan.rel	; DT_REL
	dd	.prescan.relsz	; DT_RELSZ
	dd	.prescan.relent	; DT_RELENT
	dd	.prescan.pltrel	; DT_PLTREL
	dd	.prescan.next	; DT_DEBUG
	dd	.prescan.next	; DT_TEXTREL
	dd	.prescan.jmprel	; DT_JMPREL
.prescan.jumptable_len	equ	($-.prescan.jumptable)
.prescan.jumptable_size	equ	($-.prescan.jumptable)/4

.prescan.jmprel:	; PLT Location
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR2, ecx
	jmp .prescan.next
.prescan.pltrel:	; PLT Type
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR3, ecx
	jmp .prescan.next
.prescan.pltrelsz:	; PLT Size
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR4, ecx
	jmp .prescan.next

.prescan.rela:	; RELA Table Location
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR5, ecx
	jmp .prescan.next
.prescan.relasz:	; RELA Table Size
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR6, ecx
	jmp .prescan.next
.prescan.relaent:	; RELA Table entitiy Size
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR7, ecx
	jmp .prescan.next

.prescan.rel:	; REL Table Location
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR8, ecx
	jmp .prescan.next
.prescan.relsz:	; REL Table Size
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR9, ecx
	jmp .prescan.next
.prescan.relent:	; REL Table entitiy Size
	mov ecx, DWORD [edi+tELF_Dyn.Value]
	mov DWORD VAR10, ecx
	jmp .prescan.next
	
.prescan.next:
	add edi, tELF_Dyn_size
	jmp .prescan.top
.prescan.end:

	; Let's get on to relocating

	; Plain REL entries
	mov esi, VAR8	; Check Base
	test esi, esi
	jz .REL.end
	mov edi, VAR9	; Check Size
	test edi, edi
	jz .REL.end
.REL.top:
	mov eax, [esi+tELF_Rel.Info]
	mov [esp+0], eax	; r_info
	;MAGIC_BREAK
	mov eax, [esi+tELF_Rel.Offset]
	mov [esp+4], eax	; Pointer
	mov eax, [eax]
	mov [esp+8], eax	; Addend
	mov eax, ARG1
	mov [esp+12], eax	; Base
	;DUMP_STACK	{"r_info","Pointer","*Pointer","Base"}
	call ELF_INT_DoRelocation
.REL.next:
	add esi, VAR10
	sub edi, VAR10
	jb .REL.end
	jmp .REL.top
.REL.end:


	; RELA entries
	mov esi, VAR5	; Check Base
	test esi, esi
	jz .RELA.end
	mov edi, VAR6	; Check Size
	test edi, edi
	jz .RELA.end
.RELA.top:
	mov eax, [esi+tELF_Rela.Info]
	mov [esp+0], eax	; r_info
	;MAGIC_BREAK
	mov eax, [esi+tELF_Rela.Offset]
	mov [esp+4], eax	; Pointer
	mov eax, [esi+tELF_Rela.Addend]
	mov [esp+8], eax	; Addend
	mov eax, ARG1
	mov [esp+12], eax	; Base
	;DUMP_STACK	{"r_info","Pointer","Addend","Base"}
	call ELF_INT_DoRelocation
.RELA.next:
	add esi, VAR7
	sub edi, VAR7
	jb .RELA.end
	jmp .RELA.top
.RELA.end:


	mov esi, VAR2	; Base
	test esi, esi
	jz .PLT.end
	mov edi, VAR4	; Check Size
	test edi, edi
	jz .PLT.end
	cmp DWORD VAR3, DT_REL
	jz .PLT.REL.top
; RELA
.PLT.RELA.top:
	mov eax, [esi+tELF_Rela.Info]
	mov [esp+0], eax	; r_info
	mov eax, [esi+tELF_Rela.Offset]
	mov [esp+4], eax	; Pointer
	mov eax, [esi+tELF_Rela.Addend]
	mov [esp+8], eax	; Addend
	mov eax, ARG1
	mov [esp+12], eax	; Base
	;MAGIC_BREAK
	;DEBUG	"ELF_INT_DoRelocation(r_info=0x%8,Pointer=0x%8,Addend=0x%8,Base=0x%8)",10
	call ELF_INT_DoRelocation
.PLT.RELA.next:
	;MAGIC_BREAK
	add esi, tELF_Rela_size
	sub edi, tELF_Rela_size
	jbe .PLT.end
	jmp .PLT.RELA.top
; REL
.PLT.REL.top:
	mov eax, [esi+tELF_Rel.Info]
	mov [esp+0], eax	; r_info
	mov eax, [esi+tELF_Rel.Offset]
	mov [esp+4], eax	; Pointer
	mov eax, [eax]
	mov [esp+8], eax	; Addend
	mov eax, ARG1
	mov [esp+12], eax	; Base
	;MAGIC_BREAK
	;DEBUG	"ELF_INT_DoRelocation(r_info=0x%8,Pointer=0x%8,*Pointer=0x%8,Base=0x%8)",10
	call ELF_INT_DoRelocation
.PLT.REL.next:
	;MAGIC_BREAK
	add esi, tELF_Rel_size
	sub edi, tELF_Rel_size
	jbe .PLT.end
	jmp .PLT.REL.top
.PLT.end:

	; Get Entry Point
	mov esi, ARG1
	mov eax, [esi+tELF_Header.Entrypoint]

.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; ELF_GetEntry
;; \brief Get the entrypoint
[global ELF_GetEntry]
ELF_GetEntry:
	FRAME_OPEN	0, 0, {}
	mov eax, ARG1
	mov eax, [eax+tELF_Header.Entrypoint]
	FRAME_CLOSE
	ret

;; ELF_GetSymbol
;; \brief Scan for a symbol
[global ELF_GetSymbol]
ELF_GetSymbol:
	FRAME_OPEN	0, 1, {esi,edi,ebx}
	
	mov esi, ARG1	; Base
	mov edi, [esi+tELF_Header.dynsym]
	mov esi, [esi+tELF_Header.dynhash]
	
	mov eax, ARG2	; Symbol Name
	mov [esp], eax
	call ELF_INT_GetHash
	mov ecx, [esi]	; nbuckets
	xor edx, edx
	div ecx	; EDX = EAX % ECX
	mov eax, edx	; Final Bucket number
	
	; Get Bucket value
	mov ebx, [esi+8+eax*4]
	; Get Chains pointer
	lea esi, [esi+8+ecx*4]

	;MAGIC_BREAK
.walk.top:
	test ebx, ebx
	jz .retZero
	shl ebx, 4	;	EBX *= 16
	cmp WORD [edi+ebx+tELF_Sym.Shndx], 0
	jz .walk.next
	mov eax, ARG1
	mov [esp], eax
	mov eax, [edi+ebx+tELF_Sym.Name]
	mov [esp+4], eax
	call strcmp
	jnz .walk.next
	mov eax, [edi+ebx+tELF_Sym.Value]
	mov ecx, ARG3
	mov [ecx], eax
	mov eax, 1
	jmp .ret

.walk.next:
	shr ebx, 4
	mov ebx, [esi+ebx*4]
	jmp .walk.top
.walk.end:

.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; ELF_INT_DoRelocation
;; \brief Performs a single relocation
;; \param TypeSym	r_info field of REL/RELA
;; \param Pointer	Address to perform the operation on
;; \param Addend	Base Value for operation
;; \param Base	Current binary's base
ELF_INT_DoRelocation:
	FRAME_OPEN	1, 3, {esi, edi}
	
	;MAGIC_BREAK
	; Get Symbol Name
	mov esi, ARG4	; Base
	mov esi, [esi+tELF_Header.dynsym]	; .dynsym table
	mov eax, ARG1	; Tag
	shr eax, 8	; Get Symbol
	shl eax, 4
	mov esi, [esi+eax+tELF_Sym.Name]	; Get the name
	test esi, esi
	jz .noSym
	cmp BYTE [esi], 0
	jz .noSym


	mov DWORD VAR1, 0
	; Get Symbol Value
.getSymbol:
	mov eax, ARG4
	mov [esp], eax	; Base
	mov [esp+4], esi	; Name
	lea eax, VAR1	; Value Pointer
	mov [esp+8], eax
	;DEBUG	"ELF_GetSymbol(0x%8,'%s',*0x%8)",10
	call ELF_GetSymbol
	test eax, eax
	jnz .foundSym
	
	mov [esp], esi	; Name
	lea eax, VAR1	; Value Pointer
	mov [esp+4], eax
	;DEBUG	"Binary_GetSymbol('%s',*0x%8)",10
	call Binary_GetSymbol
	jz .retZero
.foundSym:
	mov esi, VAR1
.noSym:
	
	; Get Operation
	mov eax, ARG1
	and eax, 0xFF
	cmp eax, .opr.jumptable_size
	jae	.ret
	lea eax, [.opr.jumptable+eax*4]
	jmp [eax]
.opr.jumptable:
	dd	.opr.end	;	R_386_NONE
	dd	.opr.s_a	;	R_386_32,	// S+A
	dd	.opr.s_a_p	;	R_386_PC32,	// S+A-P
	dd	.opr.unk	;.opr.g_a_p	;	R_386_GOT32,	// G+A-P
	dd	.opr.unk	;.opr.l_a_p	;	R_386_PLT32,	// L+A-P
	dd	.opr.unk	;	R_386_COPY,	// none
	dd	.opr.s		;	R_386_GLOB_DAT,	// S
	dd	.opr.s		;	R_386_JMP_SLOT,	// S
	dd	.opr.b_a	;	R_386_RELATIVE,	// B+A
	dd	.opr.unk	;.opr.s_a_got	;	R_386_GOTOFF,	// S+A-GOT
	dd	.opr.unk	;.opr.got_a_p	;	R_386_GOTPC,	// GOT+A-P
	dd	.opr.end	;	R_386_LAST	// none
.opr.jumptable_size equ ($-.opr.jumptable)/4

.opr.s_a:
	;DEBUG	"S+A",10
	mov eax, esi	; Symbol
	add eax, ARG3	; + Addend
	mov ecx, ARG2
	mov [ecx], eax
	jmp .opr.end

.opr.s_a_p:
	;DEBUG	"S+A-P",10
	mov eax, esi	; Symbol
	add eax, ARG3	; + Addend
	sub eax, ARG2	; - PC
	mov ecx, ARG2
	mov [ecx], eax
	jmp .opr.end

.opr.s:
	;DEBUG	"S",10
	mov ecx, ARG2
	mov [ecx], esi	; Symbol
	jmp .opr.end

.opr.b_a:
	;DEBUG	"B+A",10
	mov eax, ARG4	; Base
	add eax, ARG3	; + Addend
	mov ecx, ARG2
	mov [ecx], eax
	jmp .opr.end

.opr.unk:
	DEBUG	"UNK",10
	jmp .retZero
.opr.end:
	mov eax, 1
	jmp .ret

.retZero:
	;[ESP] is still the symbol name
	mov eax, ARG1
	and eax, 0xFF
	mov [esp+4], eax
	DEBUG	"Relocation Failed! (Symbol '%s', opr %i)",10,0
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; ELF_INT_GetHash
;; \brief Get a string's hash value
ELF_INT_GetHash:
	push ebp
	mov ebp, esp
	
	xor eax, eax	; Hash
	mov ecx, ARG1	; String
	
.loop:
	cmp BYTE [ecx], 0
	jz .ret
	shl eax, 4
	xor edx, edx
	mov dl, [ecx]
	add eax, edx
	mov edx, eax
	and edx, 0xF0000000
	jz .noXor
	shr edx, 24
	xor eax, edx
	shl edx, 24
.noXor:
	not edx
	and eax, edx

.next:
	inc ecx
	jmp .loop

.ret:
	pop ebp
	ret

;; ELF_INT_LoadSegment
;; \brief Called internally to load a segment from the program header table
;; \param ImageBase	Base of the image in memory
;; \param PHdrEnt	Pointer to the PHdrEnt structure
ELF_INT_LoadSegment:
	FRAME_OPEN 0, 1, {esi, edi, ebx}
	
	; Get address of structure
	mov esi, ARG2
	
	; Is this page mapped?
	mov ebx, [esi+tELF_PhEnt.VAddr]
	mov edi, [esi+tELF_PhEnt.MemSize]

.checkloop.top:
	; Check if the page is allocated
	mov [esp], ebx
	call MM_IsValid
	test eax, eax
	jnz	.checkloop.next
	; If not, allocate it
	mov eax, ebx
	and eax, 0xFFFFF000
	mov [esp], eax
	call MM_Allocate
	test eax, eax
	jz	.retOne
	
	; Go to the next page
.checkloop.next:
	add ebx, 0x1000
	cmp edi, 0x1000
	jle .checkloop.end
	sub edi, 0x1000
	jmp .checkloop.top
.checkloop.end:

	; Load data
.load:
	;xchg bx, bx
	mov edx, esi	; Save ESI
	mov edi, [esi+tELF_PhEnt.VAddr]
	mov ecx, [esi+tELF_PhEnt.FileSize]
	mov esi, [esi+tELF_PhEnt.Offset]
	add esi, ARG1
	cld
	rep movsb
	;xchg bx, bx
	mov esi, edx
	mov ecx, [esi+tELF_PhEnt.MemSize]
	sub ecx, [esi+tELF_PhEnt.FileSize]
	xor eax, eax
	rep stosb
	jmp .ret

.retOne:
	xor eax, eax
	inc eax
.ret:
	FRAME_CLOSE
	ret
