; EmulOS 2
; By thePowersgang
; binary.asm
; - Binary Loader
[BITS 32]
%include "common.inc.asm"
%include "binary.inc.asm"

[extern ELF_Load]
[extern ELF_Relocate]
[extern ELF_GetEntry]
[extern ELF_GetSymbol]
[extern Emu_GetSymbol]
[extern MM_IsValid]

[section .text]
;; Binary_LoadMem
;; \brief Loads a binary from an in-memory file image
;; \param Address	Address of the file in memory
;; \return Base address of loaded binary
[global Binary_LoadMem]
Binary_LoadMem:
	FRAME_OPEN	0, 1, {}
	
	; Pass loading on
	mov eax, ARG1
	cmp DWORD [eax], 0x7F|('E'<<8)|('L'<<16)|('F'<<24)
	jnz	.notelf
	
	; Load into memory
	mov [esp], eax
	call ELF_Load
	jmp	.ret
.notelf:
	
.retZero:
	xor eax, eax

.ret:
	FRAME_CLOSE
	ret

;; Binary_Relocate
;; \brief Relocates a loaded binary
;; \return Address of the binary's entrypoint
[global Binary_Relocate]
Binary_Relocate:
	FRAME_OPEN 0, 1, {}
	
	; Pass relocation on
	mov eax, ARG1
	
	; Check if it is an ELF
	cmp DWORD [eax], 0x7F|('E'<<8)|('L'<<16)|('F'<<24)
	jnz	.notelf
	; Relocate
	mov [esp], eax
	call ELF_Relocate
	jmp	.ret
.notelf:
	
.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; Binary_GetEntry
;; \return Address of the binary's entrypoint
[global Binary_GetEntry]
Binary_GetEntry:
	FRAME_OPEN 0, 1, {}
	
	; Pass relocation on
	mov eax, ARG1
	
	; Check if it is an ELF
	cmp DWORD [eax], 0x7F|('E'<<8)|('L'<<16)|('F'<<24)
	jnz	.notelf
	; Relocate
	mov [esp], eax
	call ELF_GetEntry
	jmp	.ret
.notelf:
	
.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; Binary_GetSymbol
;; \brief Scans all loaded binaries for a symbol
[global Binary_GetSymbol]
Binary_GetSymbol:
	FRAME_OPEN	0, 3, {esi}
	
	mov eax, ARG1
	mov [esp], eax
	mov eax, ARG2
	mov [esp+4], eax
	;DEBUG	"Emu_GetSymbol(Name='%s', Pointer=0x%8)",10
	call Emu_GetSymbol
	test eax, eax
	jnz	.ret
	
	mov eax, ARG1
	mov [esp+4], eax
	mov eax, ARG2
	mov [esp+8], eax
	
	mov esi, BIN_LOAD_LOWEST
.top:
	mov [esp], esi
	call MM_IsValid
	test eax, eax
	jz .next
	
	cmp DWORD [esi], 0x7F|('E'<<8)|('L'<<16)|('F'<<24)
	jz	.elf
	jmp .next
	
.elf:
	call ELF_GetSymbol
	test eax, eax
	jnz .ret
	jmp .next
.next:
	add esi, BIN_LOAD_QUANTUM
	cmp esi, BIN_LOAD_HIGHEST
	jl	.top

.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret
