; EmulOS 2
; By thePowersGang

; debug.asm
; - Debugging Primitives
%define _DEBUG_ASM
%include "common.inc.asm"
[BITS 32]

%macro HEXN	1
%assign bit	%1*4
%rep %1
%assign bit	bit-4
	; Nibble n
	mov eax, ecx
	shr eax, bit
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
%endrep
%endmacro

[section .text]
[global itoa]
;; itoa
;; \param Buf	Destination Buffer
;; \param Base
;; \param MinLength
;; \param Pad
;; \param Value
itoa:
	FRAME_OPEN	((65+3)/4), 0, {esi, edi, ebx}
	
	lea edi, VAR1	; Local Buffer
	mov ebx, ARG2	; Base
	xor ecx, ecx	; Index
	mov eax, ARG5	; Value (low 32-bits)

	; Lower 32-bits
.pass1.top:
	xor edx, edx
	div ebx	; EAX /= EBX, EDX = EAX % EBX
	mov dl, BYTE [cHexChars+edx]
	mov BYTE [edi+ecx], dl
	inc ecx
.pass1.next:
	test eax, eax
	jz .pass1.end
	jmp .pass1.top
.pass1.end:
	
	; Pad String
	mov al, BYTE ARG4
.pad.top:
	cmp ecx, ARG3
	jae .pad.end
.pad.next:
	mov BYTE [edi+ecx], al
	inc ecx
	jmp .pad.top
.pad.end:

	; Copy to destination buffer
	mov esi, edi
	mov edi, ARG1
.copy.top:
	dec ecx
	mov al, BYTE [esi+ecx]
	mov BYTE [edi], al
	inc edi
	test ecx, ecx
	jz .copy.end
.copy.next:
	jmp .copy.top
.copy.end:
	mov BYTE [edi], 0

.ret:
	FRAME_CLOSE
	ret

[global E9_Fmt]
E9_Fmt:
	FRAME_OPEN	((65+3)/4), 6, {esi, edi, ebx}
	
	mov esi, ARG1
	lea edi, ARG2
	mov dx, 0xE9

	lea eax, VAR1
	mov [esp], eax	; Buffer

.loop.top:
	mov al, [esi]
	test al, al
	jz .loop.end
	
	cmp al, '%'
	jz .loop.fmtcode
.loop.literal:
	out dx, al
	
	jmp .loop.next
.loop.fmtcode:
	inc esi
	mov al, [esi]
	test al, al
	jz .loop.end
	
	cmp al, '%'
	jz .loop.literal
	
	cmp al, 'X'
	jz .loop.hex16
	cmp al, '8'
	jz .loop.hex8
	cmp al, '4'
	jz .loop.hex4
	cmp al, '2'
	jz .loop.hex2
	cmp al, '1'
	jz .loop.hex1
	cmp al, 'i'
	jz .loop.sint
	cmp al, 'u'
	jz .loop.uint
	cmp al, 's'
	jz .loop.string
	jmp .loop.next

; 64-bit Hexadecimal Number
.loop.hex16:
	mov ecx, DWORD [edi+4]
	HEXN 8
	mov ecx, DWORD [edi]
	HEXN 8
	add edi, 8
	jmp .loop.next

; 8 Nibble Hexadecimal Number
.loop.hex8:
	mov ecx, DWORD [edi]
	add edi, 4
	HEXN 8
	jmp .loop.next

; 4 Nibble Hexadecimal Number
.loop.hex4:
	mov ecx, DWORD [edi]
	add edi, 4
	HEXN 4
	jmp .loop.next

; 2 Nibble Hexadecimal Number
.loop.hex2:
	mov ecx, DWORD [edi]
	add edi, 4
	HEXN 2
	jmp .loop.next

; 1 Nibble Hexadecimal Number
.loop.hex1:
	mov ecx, DWORD [edi]
	add edi, 4
	HEXN 1
	jmp .loop.next

.loop.string:
	mov ecx, DWORD [edi]
	add edi, 4
.loop.printstring:
	mov dx, 0xE9
.loop.string.top:
	mov al, BYTE [ecx]
	test al, al
	jz .loop.string.end
	out dx, al
	inc ecx
	jmp .loop.string.top
.loop.string.end:
	jmp .loop.next

; Signed Integer
.loop.sint:
	mov DWORD [esp+4], 10	; Base
	mov DWORD [esp+8], 0	; MinLength
	mov DWORD [esp+12], ' '	; Pad
	mov eax, DWORD [edi]
	add edi, 4
	test eax, 0x80000000
	jz .loop.sint.notNeg
	neg eax
.loop.sint.notNeg:
	mov DWORD [esp+16], eax	; Value
	mov DWORD [esp+20], 0	; ValueHi
	call itoa
	mov ecx, DWORD [esp]
	jmp .loop.printstring
	
; Unsigned Integer
.loop.uint:
	mov DWORD [esp+4], 10	; Base
	mov DWORD [esp+8], 0	; MinLength
	mov DWORD [esp+12], ' '	; Pad
	mov eax, DWORD [edi]
	add edi, 4
	mov DWORD [esp+16], eax	; Value
	mov DWORD [esp+20], 0	; ValueHi
	call itoa
	mov ecx, DWORD [esp]
	jmp .loop.printstring
	
.loop.integer.end:
	add edi, 4
	jmp .loop.next

.loop.next:
	inc esi
	jmp .loop.top
.loop.end:
	
	FRAME_CLOSE
	ret


[global E9_Char]
E9_Char:
	push edx
	mov dx, 0xE9
	mov al, [esp+8]
	out dx, al
	pop edx
	ret

[global E9_Str]
E9_Str:
	push ebp
	mov ebp, esp
	push edx
	push esi
	
	mov dx, 0xE9
	mov esi, [ebp+8]
.loop:
	mov al, [esi]
	test al, al
	jz .ret
	out dx, al
.ret:
	pop esi
	pop edx
	pop ebp
	ret

[global E9_Hex8]
E9_Hex8:
	push ebp
	mov ebp, esp
	push edx
	push ecx
	
	mov dx, 0xE9
	mov eax, [ebp+8]

	; Nibble 1
	mov eax, ecx
	shr eax, 28
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 2
	mov eax, ecx
	shr eax, 24
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 3
	mov eax, ecx
	shr eax, 20
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 4
	mov eax, ecx
	shr eax, 16
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 5
	mov eax, ecx
	shr eax, 12
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 6
	mov eax, ecx
	shr eax, 8
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 7
	mov eax, ecx
	shr eax, 4
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
	
	; Nibble 8
	mov eax, ecx
	and eax, 0xF
	mov al, [cHexChars+eax]
	out dx, al
.ret:
	pop ecx
	pop edx
	pop ebp
	ret

[section .rodata]
cHexChars:	db	"0123456789abcdef",0
