; EmulOS 2
; By thePowersGang
; emu/main.asm
; - Emulator Interface Core
[BITS 32]
%include "common.inc.asm"
%include "binary.inc.asm"
%include "emu/defs.inc.asm"

[extern MM_IsValid]
[extern Binary_GetEntry]
[extern Threads_Activate]
[extern TUI_Panic]
[extern TUI_Display]

EMU_DRV_MACHINE	equ	BIN_LOAD_LOWEST
EMU_DRV_CPU		equ	(BIN_LOAD_LOWEST+BIN_LOAD_QUANTUM)

[section .text]
[global Emu_Initialise]
Emu_Initialise:
	FRAME_OPEN	0, 1, {}
	
	; Check for a machine driver
	mov DWORD [esp], EMU_DRV_MACHINE
	call MM_IsValid
	test eax, eax
	jnz .haveMachineDrv
	mov DWORD [esp], csEmu_NoMachine
	call TUI_Panic
	jmp .ret
.haveMachineDrv:

	; Check for a CPU Driver
	mov DWORD [esp], EMU_DRV_CPU
	call MM_IsValid
	test eax, eax
	jnz .haveCpuDrv
	mov DWORD [esp], csEmu_NoCPU
	call TUI_Panic
	jmp .ret
.haveCpuDrv:
	
	; Initialise Machine
	mov DWORD [esp], EMU_DRV_MACHINE
	call Binary_GetEntry
	test eax, eax
	jz .ret
	; - No arguments
	call eax
	mov [gpEmu_MachineInfo], eax
	
	; Initialise CPU
	mov DWORD [esp], EMU_DRV_CPU
	call Binary_GetEntry
	test eax, eax
	jz .ret
	; Get Requested number of CPUs
	mov ecx, [gpEmu_MachineInfo]
	mov ecx, [ecx+tMachine_Definition.nCPUs]
	mov [esp], ecx
	call eax
	mov [gpEmu_CPUInfo], eax
	
	; Start CPU thread
	call Emu_StartCPUs

.ret:
	FRAME_CLOSE
	ret

;; Emu_StartCPUs
;; \brief Starts the virtual CPUs
Emu_StartCPUs:
	FRAME_OPEN	0, 4, {}
	
	;TODO: Start multiple VCPUs
	mov DWORD [esp], PROC_THREAD_VCPU0
	mov DWORD [esp+4], Emu_UpdateCPU
	mov DWORD [esp+8], 1
	mov DWORD [esp+12], 0	; VCPU Num
	call Threads_Activate
	
	FRAME_CLOSE
	ret

Emu_UpdateCPU:
	sub esp, 4*4	; We don't really have a proper stack frame here
	
	mov edi, [gpEmu_CPUInfo]	; Get the CPU Info Structure
	mov esi, [edi+tCPU_Definition.Update]	; Get the Update Function
	
	mov eax, [esp+4*4+4]	; Argument (CPU Number)
	mov ecx, [edi+tCPU_Definition.CPUs]	; CPU Base
	mov eax, [ecx+eax*4]
	mov [esp], eax	; CPU Structure
.loop:
	;call Emu_DumpRegs

	call esi
	test eax, eax
	jnz .err
	hlt
	jmp .loop
.err:
	mov DWORD [esp], 0	; X
	mov DWORD [esp+4], 24	; Y
	mov DWORD [esp+8], csEmu_CPUError
	mov DWORD [esp+12], eax
	call TUI_Display
.hlt:
	hlt
	jmp .hlt

[global Emu_DumpRegs]
Emu_DumpRegs:
	FRAME_OPEN	0, 7, {esi,edi,ebx}
	
	mov eax, ARG1
	mov [esp], eax
	mov esi, [gpEmu_CPUInfo]	; Get the CPU Info Structure
	mov ecx, [esi+tCPU_Definition.GetRegs]
	call ecx
	mov ebx, [esi+tCPU_Definition.nRegs]
	mov edi, [esi+tCPU_Definition.Regs]
	mov esi, eax
	
	mov DWORD [esp], csEmu_Dump64Bits
.loop:	
	mov edx, DWORD [edi+0]
	mov DWORD [esp+4], edx
	mov edx, DWORD [esi+4]
	mov DWORD [esp+8], edx
	mov edx, DWORD [esi+0]
	mov DWORD [esp+12], edx
	
	mov edx, DWORD [edi+4]
	mov DWORD [esp+16], edx
	mov edx, DWORD [esi+12]
	mov DWORD [esp+20], edx
	mov edx, DWORD [esi+8]
	mov DWORD [esp+24], edx
	call E9_Fmt
	
	add esi, 8*2
	add edi, 4*2
	sub ebx, 2
	jbe .ret
	jmp .loop

.ret:
	FRAME_CLOSE
	ret

[section .bss]
gpEmu_MachineInfo:	resd	0
gpEmu_CPUInfo:	resd	0

[section .rodata]
csEmu_Dump64Bits:
	db	"%s %8:%8 %s %8:%8",10,0
csEmu_NoMachine:
	db	"No machine driver loaded",0
csEmu_NoCPU:
	db	"No CPU driver loaded",0
csEmu_CPUError:
	db	"CPU Returned Error: %x",0
