; EmulOS 2
; By thePowersGang
; emu/mem.asm
; - Emulated Memory Management/Control
[BITS 32]
%include "common.inc.asm"
%include "emu/memory.inc.asm"

; === CONSTANTS ===

; === VARIABLES ===
[section .data]
giMemory_NextBusLockID:	dd	1
[section .bss]
giMemory_BusLockID:	resd	1
gEmuMem_Regions:	resd	1
gbMemory_IsBusLocked:	resd	1

; === CODE ===
[section .text]
;; Memory_Register
;; \brief Registers a memory region
;; \param Region	Pointer to the region structure
[global Memory_Register]
Memory_Register:
	push ebp
	mov ebp, esp
	push esi
	push ebx
	
	mov esi, ARG1
	xor ebx, ebx
	
	mov eax, [gEmuMem_Regions]
	test eax, eax
	jz	.onlyOne

	; Find a gap
.findPrev.top:
	; Check high components
	mov ecx, [eax+tEmuMemRegion.Base+4]
	cmp ecx, [esi+tEmuMemRegion.Base+4]
	jl .foundGap
	jg .findPrev.next
	mov ecx, [eax+tEmuMemRegion.Base]
	cmp ecx, [esi+tEmuMemRegion.Base]
	jl .foundGap
	jz .retZero
	; Greater?, NEXT!

.findPrev.next:
	mov ebx, eax	; Save last
	mov eax, [eax+tEmuMemRegion.Next]
	test eax, eax
	jz .addToEnd
	jmp .findPrev.top

.foundGap:
	; EBX - Previous (or zero for before first)
	; EAX - Next
	; ESI - New
	mov ecx, eax
	test ebx, ebx
	jz .addToFront
	
	; Check for overlap with previous
	mov eax, [esi+tEmuMemRegion.Base]
	mov edx, [esi+tEmuMemRegion.Base+4]
	sub eax, [ebx+tEmuMemRegion.Base]
	sub edx, [ebx+tEmuMemRegion.Base+4]
	cmp eax, [ebx+tEmuMemRegion.Length]
	jg .foundGap.maybeSafe
	cmp edx, [ebx+tEmuMemRegion.Length+4]
	jl .retZero	; Uh-oh, they overlap!
.foundGap.maybeSafe:
	; Check for overlap with next
	mov eax, [ecx+tEmuMemRegion.Base]
	mov edx, [ecx+tEmuMemRegion.Base+4]
	sub eax, [esi+tEmuMemRegion.Base]
	sub edx, [esi+tEmuMemRegion.Base+4]
	cmp eax, [esi+tEmuMemRegion.Length]
	jg .foundGap.safe
	cmp edx, [esi+tEmuMemRegion.Length+4]
	jl .retZero	; An overlap!
.foundGap.safe:
	; Insert into list
	mov [esi+tEmuMemRegion.Next], ecx
	mov [ebx+tEmuMemRegion.Next], esi
	jmp .retOne

.addToFront:
	mov eax, [ecx+tEmuMemRegion.Base]
	mov edx, [ecx+tEmuMemRegion.Base+4]
	sub eax, [esi+tEmuMemRegion.Base]
	sub edx, [esi+tEmuMemRegion.Base+4]
	cmp eax, [esi+tEmuMemRegion.Length]
	jg .addToFront.safe
	cmp edx, [esi+tEmuMemRegion.Length+4]
	jl .retZero	; An overlap!
.addToFront.safe:
	mov [esi+tEmuMemRegion.Next], ecx
	mov [gEmuMem_Regions], esi
	jmp .retOne

.addToEnd:
	; EBX - end of list
	mov eax, [esi+tEmuMemRegion.Base]
	mov edx, [esi+tEmuMemRegion.Base+4]
	sub eax, [ebx+tEmuMemRegion.Base]
	sub edx, [ebx+tEmuMemRegion.Base+4]
	cmp eax, [ebx+tEmuMemRegion.Length]
	jg .addToEnd.safe
	cmp edx, [ebx+tEmuMemRegion.Length+4]
	jl .retZero	; Uh-oh, they overlap!
.addToEnd.safe:
	mov [ebx+tEmuMemRegion.Next], esi
	mov DWORD [esi+tEmuMemRegion.Next], 0
	jmp .retOne
	
.onlyOne:
	mov [gEmuMem_Regions], esi
	jmp .retOne

.retZero:
	xor eax, eax
	jmp .ret
.retOne:
	xor eax, eax
	inc eax
.ret:
	pop ebx
	pop esi
	pop ebp
	ret


;; Memory_Read
;; \brief Read from emulated memory
;; \param AtomicHandle
;; \param AddressLow	Emulated physical address
;; \param AddressHigh	Emulated physical address (high bits)
;; \param Size	Size of read
;; \param Destination	Destination for read bytes
;; \return Boolean Success
[global Memory_Read]
Memory_Read:
	FRAME_OPEN	0, 5, {}
	
	; Check for atomic access
	cmp BYTE [gbMemory_IsBusLocked], 0
	jz .gotbus
	mov eax, ARG1
.waitforbus:
	cmp DWORD [giMemory_BusLockID], eax
	jz .gotbus
	hlt
	jmp .waitforbus
.gotbus:
	
	mov eax, ARG2	; Address Low
	mov [esp], eax
	mov eax, ARG3	; Address High
	mov [esp+4], eax
	;DEBUG	"Memory_INT_GetRegion(0x%X)"
	call	Memory_INT_GetRegion
	mov [esp], eax
	;DEBUG	" = *0x%8",10
	test eax, eax
	jz .retZero
	; EAX is now the region
	
	; Check if the callback is valid
	mov ecx, [eax+tEmuMemRegion.Read]
	test ecx, ecx
	jz	.retZero
	
	mov [esp], eax	; Stucture
	
	mov edx, ARG2
	sub edx, [eax+tEmuMemRegion.Base]
	mov [esp+4], edx	; Address Low (Relative)
	
	mov edx, ARG3
	sub edx, [eax+tEmuMemRegion.Base+4]
	mov [esp+8], edx	; Address High (Relative)
	
	mov edx, ARG4	; Size
	mov [esp+12], edx
	mov edx, ARG5	; Destination
	mov [esp+16], edx
	;DEBUG	"{ecx}(*0x%8, 0x%X, %i, *0x%8)",10
	call ecx	; Call function
	jmp .ret

.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; Memory_Write
;; \brief Write to emulated memory
;; \param AtomicHandle
;; \param Address	Emulated physical address
;; \param Size	Size of write
;; \param Source	Source for written bytes
;; \return Boolean Success
[global Memory_Write]
Memory_Write:
	FRAME_OPEN	0, 5, {}
	
	cmp BYTE [gbMemory_IsBusLocked], 0
	jz .gotbus
	mov eax, ARG1
.waitforbus:
	lock cmp DWORD [giMemory_BusLockID], eax
	jz .gotbus
	hlt
	jmp .waitforbus
.gotbus:
	
	mov eax, ARG2	; Address Low
	mov [esp], eax
	mov eax, ARG3	; Address High
	mov [esp+4], eax
	call	Memory_INT_GetRegion
	test eax, eax
	jz .retZero
	; EAX is now the region
	
	mov ecx, [eax+tEmuMemRegion.Write]
	test ecx, ecx
	jz	.retZero
	
	mov [esp], eax	; Stucture
	
	mov edx, ARG2
	sub edx, [eax+tEmuMemRegion.Base]
	mov [esp+4], edx	; Address Low (Relative)
	
	mov edx, ARG3
	sub edx, [eax+tEmuMemRegion.Base+4]
	mov [esp+8], edx	; Address High (Relative)
	
	mov edx, ARG4	; Size
	mov [esp+12], edx
	mov edx, ARG5	; Destination
	mov [esp+16], edx
	call ecx	; Call function
	jmp .ret
	
.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; Memory_INT_GetRegion
;; \brief Gets the region structure associated with an addres
;; \param AddressLow
;; \param AddressHi
Memory_INT_GetRegion:
	FRAME_OPEN	0, 1, {ebx}
	
	mov eax, [gEmuMem_Regions]

.top:
	mov [esp], eax
	;DEBUG	" Memory_INT_GetRegion: Region 0x%8",10
	test eax, eax
	jz .retZero
	
	; Load Arguments
	mov ecx, ARG1	; Low Address
	mov ebx, ARG2	; High Address
	
	mov [esp], ebx
	mov edx, [eax+tEmuMemRegion.Base+4]
	mov [esp+4], edx
	;DEBUG	" Memory_INT_GetRegion: cmp ebx(0x%8), [eax+tEmuMemRegion.Base+4](%8)",10
	; Check base (if lower than base, return zero)
	cmp ebx, [eax+tEmuMemRegion.Base+4]
	jb .retZero
	mov [esp], ecx
	mov edx, [eax+tEmuMemRegion.Base]
	mov [esp+4], edx
	;DEBUG	" Memory_INT_GetRegion: cmp ecx(0x%8), [eax+tEmuMemRegion.Base](%8)",10
	cmp ecx, [eax+tEmuMemRegion.Base]
	jb .retZero
	
	; Get offset in region
	sub ecx, [eax+tEmuMemRegion.Base]	; Sub with carry
	sbb ebx, [eax+tEmuMemRegion.Base+4]
	
	mov [esp], ecx
	mov [esp+4], ebx
	;DEBUG	" Memory_INT_GetRegion: Offset = %X",10
	
	; Check high bits
	cmp ebx, [eax+tEmuMemRegion.Length+4]
	ja	.next	; If this is larger than the region, skip
	
	; Check low bits
	cmp ecx, [eax+tEmuMemRegion.Length]
	jb .ret
.next:
	mov eax, [eax+tEmuMemRegion.Base]
	jmp .top

.retZero:
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; Memory_LockBus
;; \brief Locks the virtual address bus
;; \return Lock ID
[global Memory_LockBus]
Memory_LockBus:
	mov eax, 1
.spinlock:
	lock cmpxchg [gbMemory_IsBusLocked], eax
	cmp eax, 0
	jz .out
	hlt
	jmp .spinlock
.out:

	mov eax, [giMemory_NextBusLockID]
	inc DWORD [giMemory_NextBusLockID]
	mov [giMemory_BusLockID], eax
.ret:
	ret

;; Memory_UnlockBus
;; \brief Locks the virtual address bus
;; \param Lock ID
[global Memory_UnlockBus]
Memory_UnlockBus:
	mov eax, [esp+4]
	cmp [giMemory_BusLockID], eax
	jnz .ret
	mov BYTE [gbMemory_IsBusLocked], 0
	mov DWORD [giMemory_BusLockID], 0
.ret:
	ret
