; EmulOS 2
; By thePowersGang
; idt.asm
%include "common.inc.asm"

[extern TUI_Display]

[BITS 32]
[section .data]
ALIGN 8
gIDT:
	times	256	dd	0x00080000,0x00000F00
gIDTPtr:
	dw	256 * 16 - 1	; Limit
	dd	gIDT		; Base


[section .text]

[global IDT_Initialise]
IDT_Initialise:

	; Set IDT
%macro SETISR	1
	mov eax, Isr%1
	mov	WORD [gIDT + %1*8], ax
	shr eax, 16
	mov	WORD [gIDT + %1*8+6], ax
	mov	ax, WORD [gIDT + %1*8 + 4]
	or ax, 0x8000
	mov	WORD [gIDT + %1*8 + 4], ax
%endmacro
	
	; ISRs
	%assign	i	0
	%rep 32
	SETISR	i
	%assign i i+1
	%endrep
	
	; IRQs
	%assign	i	0xF0
	%rep 16
	SETISR	i
	%assign i i+1
	%endrep
	
	; Load IDT
	lidt [gIDTPtr]

	; Remap PIC
	push edx	; Save EDX
	mov dx, 0x20
	mov al, 0x11
	out dx, al	;	Init Command
    mov dx, 0x21
	mov al, 0xF0
	out dx, al	;	Offset (Start of IDT Range)
    mov al, 0x04
	out dx, al	;	IRQ connected to Slave (00000100b) = IRQ2
    mov al, 0x01
	out dx, al	;	Set Mode
    mov al, 0x00
	out dx, al	;	Set Mode
	
	mov dx, 0xA0
	mov al, 0x11
	out dx, al	;	Init Command
    mov dx, 0xA1
	mov al, 0xF8
	out dx, al	;	Offset (Start of IDT Range)
    mov al, 0x02
	out dx, al	;	IRQ Line connected to master
    mov al, 0x01
	out dx, al	;	Set Mode
    mov dl, 0x00
	out dx, al	;	Set Mode
	pop edx
	
	ret

%macro ISR_NERRNO	1
Isr%1:
	;xchg bx, bx
	push 0
	push %1
	jmp	IsrCommon
%endmacro
%macro ISR_ERRNO	1
Isr%1:
	;xchg bx, bx
	push %1
	jmp	IsrCommon
%endmacro

%macro DEF_IRQ	1
[global Isr%1]
Isr%1:
	push	0
	push	%1
	jmp	IRQCommon
%endmacro

[extern TimerIRQCommon]
; IRQs
%if CPU0_USES_PIT
; - Timer
[global Isr240]
Isr240:
	push	0	; CPU 0
	jmp TimerIRQCommon
%else
	DEF_IRQ	240
%endif
; - Assignable
%assign i	0xF1
%rep 16
	DEF_IRQ	i
%assign i i+1
%endrep

ISR_NERRNO	0;  0: Divide By Zero Exception
ISR_NERRNO	1;  1: Debug Exception
ISR_NERRNO	2;  2: Non Maskable Interrupt Exception
ISR_NERRNO	3;  3: Int 3 Exception
ISR_NERRNO	4;  4: INTO Exception
ISR_NERRNO	5;  5: Out of Bounds Exception
ISR_NERRNO	6;  6: Invalid Opcode Exception
ISR_NERRNO	7;  7: Coprocessor Not Available Exception
ISR_ERRNO	8;  8: Double Fault Exception (With Error Code!)
ISR_NERRNO	9;  9: Coprocessor Segment Overrun Exception
ISR_ERRNO	10; 10: Bad TSS Exception (With Error Code!)
ISR_ERRNO	11; 11: Segment Not Present Exception (With Error Code!)
ISR_ERRNO	12; 12: Stack Fault Exception (With Error Code!)
ISR_ERRNO	13; 13: General Protection Fault Exception (With Error Code!)
ISR_ERRNO	14; 14: Page Fault Exception (With Error Code!)
ISR_NERRNO	15; 15: Reserved Exception
ISR_NERRNO	16; 16: Floating Point Exception
ISR_NERRNO	17; 17: Alignment Check Exception
ISR_NERRNO	18; 18: Machine Check Exception
ISR_NERRNO	19; 19: Reserved
ISR_NERRNO	20; 20: Reserved
ISR_NERRNO	21; 21: Reserved
ISR_NERRNO	22; 22: Reserved
ISR_NERRNO	23; 23: Reserved
ISR_NERRNO	24; 24: Reserved
ISR_NERRNO	25; 25: Reserved
ISR_NERRNO	26; 26: Reserved
ISR_NERRNO	27; 27: Reserved
ISR_NERRNO	28; 28: Reserved
ISR_NERRNO	29; 29: Reserved
ISR_NERRNO	30; 30: Reserved
ISR_NERRNO	31; 31: Reserved

IsrCommon:
	xchg bx, bx
	pusha
	mov ebp, esp
	sub esp, 7*4
	
	; Print Reg Dump
	; Line 0
	mov eax, [ebp+8*4]	; IntNum
	mov DWORD [esp], 0	; X
	mov DWORD [esp+4], 8	; Y
	mov DWORD [esp+8], csIDT_Line0
	mov eax, DWORD [csIDT_ErrorNames+eax*4]
	mov DWORD [esp+12], eax	; Error Name
	call TUI_Display
	
	; Line 1
	inc DWORD [esp+4]
	mov DWORD [esp+8], csIDT_Line1
	mov eax, cr2
	mov DWORD [esp+12], eax	; CR2
	call TUI_Display
	
	; Line 2
	inc DWORD [esp+4]
	mov DWORD [esp+8], csIDT_Line2
	mov eax, [ebp+8*4+8]	; EIP
	mov DWORD [esp+12], eax
	lea eax, [ebp+8*4+12]	; ESP
	mov DWORD [esp+16], eax
	call TUI_Display
	
	; Line 3
	inc DWORD [esp+4]
	mov DWORD [esp+8], csIDT_Line3
	mov eax, [ebp+8*4-4]	; EAX
	mov DWORD [esp+12], eax
	mov eax, [ebp+8*4-8]	; ECX
	mov DWORD [esp+16], eax
	mov eax, [ebp+8*4-12]	; EDX
	mov DWORD [esp+20], eax
	mov eax, [ebp+8*4-16]	; EBX
	mov DWORD [esp+24], eax
	call TUI_Display
	
	; Line 4
	inc DWORD [esp+4]
	mov DWORD [esp+8], csIDT_Line4
	mov eax, [ebp+8*4-20]	; ESP
	mov DWORD [esp+12], eax
	mov eax, [ebp+8*4-24]	; EBP
	mov DWORD [esp+16], eax
	mov eax, [ebp+8*4-28]	; ESI
	mov DWORD [esp+20], eax
	mov eax, [ebp+8*4-32]	; EDI
	mov DWORD [esp+24], eax
	call TUI_Display
	
	add esp, 7*4
	cli
	hlt
	jmp $

IRQCommon:
	add esp, 8
	iret

[section .rodata]
csIDT_Line0:
	db	"CPU Fault: %s",0
csIDT_Line1:
	db	"CR2 %x",0
csIDT_Line2:
	db	"EIP %x ESP %x",0
csIDT_Line3:
	db	"EAX %x ECX %x EDX %x EBX %x",0
csIDT_Line4:
	db	"ESP %x EBP %x ESI %x EDI %x",0
csIDT_ErrorNames:
	dd	csIDT_ErrorName_DIV0,	csIDT_ErrorName_Debug	; 0, 1
	dd	csIDT_ErrorName_NMI,	csIDT_ErrorName_Breakpoint	; 2, 3
	dd	0, 0	; 4, 5
	dd	csIDT_ErrorName_InvlOpcode,	csIDT_ErrorName_NoCoproc	; 6, 7
	dd	csIDT_ErrorName_DoubleFault,	0	; 8, 9
	dd	csIDT_ErrorName_BadTSS,	0	; 10, 11
	dd	csIDT_ErrorName_StackFault, csIDT_ErrorName_GPF	; 12, 13
	dd	csIDT_ErrorName_PageFault,	csIDT_ErrorName_UnkInt	; 14, 15
	dd	csIDT_ErrorName_CoprocFault,	csIDT_ErrorName_AlignCheck	; 16, 17
	dd	csIDT_ErrorName_MachineCheck,	csIDT_ErrorName_Reserved	; 18, 19
	dd	csIDT_ErrorName_Reserved,	csIDT_ErrorName_Reserved	; 20, 21
	dd	csIDT_ErrorName_Reserved,	csIDT_ErrorName_Reserved	; 22, 23
	dd	csIDT_ErrorName_Reserved,	csIDT_ErrorName_Reserved	; 24, 25
	dd	csIDT_ErrorName_Reserved,	csIDT_ErrorName_Reserved	; 26, 27
	dd	csIDT_ErrorName_Reserved,	csIDT_ErrorName_Reserved	; 28, 29
	dd	csIDT_ErrorName_Reserved,	csIDT_ErrorName_Reserved	; 30, 31

csIDT_ErrorName_DIV0:
	db	"Division By Zero",0
csIDT_ErrorName_Debug:
	db	"Debug Exception",0
csIDT_ErrorName_NMI:
	db	"Non-Maskable Interrupt",0
csIDT_ErrorName_Breakpoint:
	db	"Breakpoint",0

csIDT_ErrorName_InvlOpcode:
	db	"Invalid Opcode",0
csIDT_ErrorName_NoCoproc:
	db	"No Coprocessor",0
csIDT_ErrorName_DoubleFault:
	db	"Double Fault",0

csIDT_ErrorName_BadTSS:
	db	"Bad TSS",0

csIDT_ErrorName_StackFault:
	db	"Stack Fault",0
csIDT_ErrorName_GPF:
	db	"General Protection Fault",0
csIDT_ErrorName_PageFault:
	db	"Page Fault",0
csIDT_ErrorName_UnkInt:
	db	"Unknown Interrupt",0

csIDT_ErrorName_CoprocFault:
	db	"Coprocessor Fault",0
csIDT_ErrorName_AlignCheck:
	db	"Alignment Check",0
csIDT_ErrorName_MachineCheck:
	db	"Machine Check",0
csIDT_ErrorName_Reserved:
	db	"Reserved",0
