; EmulOS 2
; By thePowersGang
%include "common.inc.asm"
%include "mboot.inc.asm"
%include "binary.inc.asm"
[BITS 32]
[section .text]

[extern TUI_ClearScreen]
[extern IDT_Initialise]
[extern MM_Initialise]
[extern Threads_Init]
[extern Emu_Initialise]
[extern TUI_Display]
[extern Binary_LoadMem]
[extern Binary_Relocate]
[extern MM_IsValid]
[extern _giMultibootPtr]

[global main]
main:
	sub esp, 3*4
	call	TUI_ClearScreen
	
	; Set up IDT
	call	IDT_Initialise

	; Set up Memory Manager
	call	MM_Initialise

	; Start Threads
	call	Threads_Init

	;cli
	
	mov DWORD [esp], 80/2-csLoadedString.len/2
	mov DWORD [esp+4], 0
	mov DWORD [esp+8], csLoadedString
	call TUI_Display

	call LoadModules
	call RelocateModules
	
	call Emu_Initialise

.ret:
	add esp, 3*4
.hlt:
	hlt
	jmp .hlt

;; LoadModules
;; \brief Loads Mutiboot modules
LoadModules:
	push ebp
	mov ebp, esp
	push edi
	push esi
	push ebx
	sub esp, 5*4

	mov esi, [_giMultibootPtr]
	mov ebx, [esi+tMBoot.ModuleCount]
	test ebx, ebx
	jz .ret
	
	add DWORD [esi+tMBoot.FirstModule], KERNEL_BASE
	mov edi, [esi+tMBoot.FirstModule]
	
	mov DWORD [esp+4], 1

.load.top:
	mov DWORD [esp], 0	; This will get clobbered when we call Binary_LoadMem
	; Row is a Stack Constant
	mov DWORD [esp+8], csPrintModule
	
	mov eax, DWORD [edi+tMBootModule.Start]
	mov [esp+12], eax
	mov eax, DWORD [edi+tMBootModule.String]
	add eax, KERNEL_BASE
	mov [esp+16], eax
	
	call TUI_Display
	
	; Disaster waiting to happen if the module is loaded past 4MiB
	mov eax, DWORD [edi+tMBootModule.Start]
	add eax, KERNEL_BASE
	mov DWORD [esp], eax
	call Binary_LoadMem
	
	mov DWORD [esp+0], 80-(csPrintModuleLoaded.len-2+8)
	mov DWORD [esp+8], csPrintModuleLoaded
	mov DWORD [esp+12], eax
	call TUI_Display
	
.load.next:
	dec ebx
	jz .load.end
	inc DWORD [esp+4]
	add edi, tMBootModule_size
	jmp .load.top
.load.end:

.ret:
	add esp, 5*4
	pop ebx
	pop esi
	pop edi
	pop ebp
	ret

RelocateModules:
	push ebp
	mov ebp, esp
	push esi
	push ebx
	sub esp, 1*4
	
	xor ebx, ebx
	mov esi, BIN_LOAD_LOWEST
.top:
	mov [esp], esi
	call MM_IsValid
	test eax, eax
	jz .next
	
	call Binary_Relocate
	mov [gaModuleEntries+ebx*4], eax
	
.next:
	inc ebx
	add esi, BIN_LOAD_QUANTUM
	cmp esi, BIN_LOAD_HIGHEST
	jl .top
	
.ret:
	add esp, 1*4
	pop ebx
	pop ebp
	pop esi
	ret

[section .data]
gaModuleEntries:
	times (BIN_LOAD_HIGHEST-BIN_LOAD_LOWEST)/BIN_LOAD_QUANTUM	dd	0

[section .rodata]
csPrintModule:
	db	"Module: %x '%s'",0
csPrintModule.len equ $-csPrintModule-1

csPrintModuleLoaded:
	db	"Loaded at %x",0
csPrintModuleLoaded.len equ $-csPrintModuleLoaded-1

csLoadedString:
	db	"EmulOS 2 Loaded",0
csLoadedString.len	equ $-csLoadedString-1
