; EmulOS 2
; By thePowersGang
; mm.asm
; - Physical and Virtual Memory Manager
[BITS 32]
%include "common.inc.asm"
%include "mboot.inc.asm"

; === CONSTANTS ===
%define MAX_PAGES	(1024*1024)
MM_FRACTAL_DIR	equ	(MM_FRACTAL+(MM_FRACTAL>>10))

; === IMPORTS ===
; link.ld
[extern start]
[extern end]
; start.asm
[extern _giMultibootPtr]
[extern _gMainPageDir]
[extern _gKernelPageTable]

; === VARIABLES ===
[section .bss]
gaMM_SuperBitmap:
	resb	(MAX_PAGES+31)/(32*32)	; Each bit is a used DW in gMM_Bitmap
gaMM_Bitmap:
	resd	(MAX_PAGES+31)/32

[section .data]
gMM_LastUsedPage:	dd	0	; Last unsure page
gMM_LastValidPage:	dd	0	; Last Addressable Page

; === CODE ===
[section .text]
;; MM_Initialise
;; \brief Initialise the Memory Manager
[global MM_Initialise]
MM_Initialise:
	FRAME_OPEN	0, 2, {edi, esi}
	
	; Start clearing according to the provided memory map
	mov esi, [_giMultibootPtr]
	mov ecx, [esi+tMBoot.MemMapSize]
	mov esi, [esi+tMBoot.MemMapAddr]
	add esi, KERNEL_BASE
	
	; Check for an empty memory map
	test ecx, ecx
	jz	.usenaive
	
	; Initialise the bitmap to one
	cld
	mov ecx, (MAX_PAGES+31)/(32*32)
	mov edi, gaMM_SuperBitmap
	mov eax, 0xFFFFFFFF
	rep stosd
	mov ecx, MAX_PAGES/32
	mov edi, gaMM_Bitmap
	rep stosd
	
	; Read the memory map
.readmap.top:
	cmp DWORD [esi+tMBootMap.Type], 1
	jnz .readmap.next
	
	mov eax, [esi+tMBootMap.Base]
	mov DWORD [esp], eax
	mov eax, [esi+tMBootMap.Length]
	mov DWORD [esp+4], eax
	call MM_INT_ClearRangeInBitmap
	
	mov eax, [esi+tMBootMap.Base]
	add eax, [esi+tMBootMap.Length]
	cmp eax, [gMM_LastValidPage]
	jbe .readmap.next
	mov [gMM_LastValidPage], eax
	
.readmap.next:
	sub ecx, [esi+tMBootMap.Size]
	jz	.readmap.out
	mov edi, [esi+tMBootMap.Size]
	add esi, 4
	jmp .readmap.top
.readmap.out:
	jmp .fillUs
	
.usenaive:
	mov DWORD [esp], 0
	mov DWORD [esp+4], 0x100000
	call MM_INT_SetRangeInBitmap
	
.fillUs:
	or DWORD [gaMM_Bitmap], 1

	mov DWORD [esp], 0x100000
	mov DWORD [esp+4], 0x100000	; Well, 1MiB should be enough for now (until I get more intelligent marking)
	;xchg bx, bx
	call MM_INT_SetRangeInBitmap

.setupVirtual:
	; Set up virtual address space
	mov DWORD [_gMainPageDir+0], 0
	mov DWORD [_gMainPageDir+(KERNEL_BASE>>22)*4], _gKernelPageTable-KERNEL_BASE+3
	mov DWORD [_gMainPageDir+(MM_FRACTAL>>22)*4], _gMainPageDir-KERNEL_BASE+3
	invlpg [0]
	invlpg [MM_FRACTAL]
	
.ret:
	FRAME_CLOSE
	ret

;; MM_AllocPhys
;; \brief Allocates a physical page
MM_AllocPhys:
	FRAME_OPEN	0, 0, {ebx}
	
	mov eax, [gMM_LastUsedPage]

.scan.top:
	;xchg bx, bx
	cmp eax, [gMM_LastValidPage]
	jae .retZero
	mov ecx, eax
	shr ecx, 5+5
	mov edx, [gaMM_SuperBitmap+ecx*4]
	inc edx
	jz	.scan.jump1024
	
	mov ecx, eax
	shr ecx, 5
	mov edx, [gaMM_Bitmap+ecx*4]
	inc edx
	jz .scan.jump32
	dec edx
	
	mov ecx, eax
	and ecx, 31
	mov ebx, 1
	shl ebx, cl
	test edx, ebx
	jz .foundOne
	
	inc eax
	jmp .scan.top

.scan.jump1024:
	add eax, 1024
	jmp .scan.top
.scan.jump32:
	add eax, 32
	jmp .scan.top
.scan.end:
	
.foundOne:
	mov [gMM_LastUsedPage], eax
	; EBX is still the mask for this
	mov ecx, eax
	shr ecx, 12
	or [gaMM_Bitmap+ecx*4], ebx
	mov ebx, [gaMM_Bitmap+ecx*4]
	inc ebx
	jnz .superNotFull
	; Ok, the superpage is full :(
	mov ecx, eax
	shr ecx, 12
	and ecx, 31
	mov ebx, 1
	shr ebx, cl
	mov ecx, eax
	shr ecx, 22
	or [gaMM_SuperBitmap+ecx*4], ebx
	
.superNotFull:
	shl eax, 12
	;xchg bx, bx
	jmp .ret
.retZero:
	xchg bx, bx
	xor eax, eax
.ret:
	FRAME_CLOSE
	ret

;; MM_IsValid
;; \brief Checks if a page is accessable
;; \param Address	Virtual Address to check
[global MM_IsValid]
MM_IsValid:
	push ebp
	mov ebp, esp
	
	; Test Directory
	mov eax, DWORD ARG1
	shr eax, 22
	mov eax, [MM_FRACTAL_DIR+eax*4]
	test eax, 1
	jz	.retZero
	
	; Test Table
	mov eax, DWORD ARG1
	shr eax, 12
	mov eax, [MM_FRACTAL+eax*4]
	and eax, 1
	jnz	.ret

.retZero:
	xor eax, eax
.ret:
	pop ebp
	ret

;; MM_Allocate
;; \brief Allocate memory at a virtual address
;; \param Address	Virtual Address
;; \return Allocated Physical address or Zero
[global MM_Allocate]
MM_Allocate:
	push ebp
	mov ebp, esp
	push ebx
	push edi
	
	; Check if there is a directory mapped
	mov ebx, DWORD ARG1
	shr ebx, 22
	mov eax, [MM_FRACTAL_DIR+ebx*4]
	test eax, eax
	jnz	.dirMapped

	; Allocate one
	call MM_AllocPhys
	test eax, eax
	jz	.ret
	or eax, 3
	mov [MM_FRACTAL_DIR+ebx*4], eax
	; Clear it
	mov edi, ebx
	shl edi, 12
	add edi, MM_FRACTAL
	invlpg [edi]
	mov ecx, 1024
	xor eax, eax
	rep stosd

.dirMapped:
	mov ebx, DWORD ARG1
	shr ebx, 12
	mov eax, [MM_FRACTAL+ebx*4]
	test eax, eax
	jnz	.ret	; Already mapped
	
	call MM_AllocPhys
	test eax, eax
	jz	.ret
	or eax, 3
	mov [MM_FRACTAL+ebx*4], eax
	
	and eax, ~0xFFF
	
	shl ebx, 12
	invlpg [ebx]
.ret:
	pop edi
	pop ebx
	pop ebp
	ret

;; MM_INT_ClearRangeInBitmap
;; \brief Clears a range in gaMM_Bitmap
;; \param Base	Base of memory range
;; \param NumBytes	Number of bytes in range
MM_INT_ClearRangeInBitmap:
	push ebp
	mov ebp, esp
	push edi
	sub esp, 8
	
	mov edi, DWORD ARG1
	mov ecx, DWORD ARG2
	add ecx, 0xFFF
	shr ecx, 12	;	ECX /= 4096
	test ecx, ecx
	jz	.ret
	
	;xchg bx, bx
	
	mov DWORD [esp], edi	; Page
	mov DWORD [esp+4], 0
	
.loop:
	call	MM_INT_MarkAs
	add DWORD [esp], 0x1000
	dec ecx
	jnz .loop

.ret:
	add esp, 8
	pop edi
	pop ebp
	ret

;; MM_INT_SetRangeInBitmap
;; \brief Sets a range in gaMM_Bitmap
;; \param Base	Base of memory range
;; \param NumBytes	Number of bytes in range
MM_INT_SetRangeInBitmap:
	push ebp
	mov ebp, esp
	push edi
	sub esp, 8
	
	mov edi, DWORD ARG1
	mov ecx, DWORD ARG2
	add ecx, 0xFFF
	shr ecx, 12	;	ECX /= 4096
	
	mov DWORD [esp], edi	; Page
	mov DWORD [esp+4], 1

.loop:
	call MM_INT_MarkAs
	add DWORD [esp], 0x1000
	dec ecx
	jnz .loop
	
	add esp, 8
	pop edi
	pop ebp
	ret

;; MM_INT_MarkAs
;; \brief Marks a page as either used or unused in the bitmap
MM_INT_MarkAs:
	push ebp
	mov ebp, esp
	push edi
	
	mov ecx, DWORD ARG1
	shr ecx, 12
	mov edi, ecx
	shr edi, 5	; / 32
	and ecx, 31
	
	mov eax, 1
	shl eax, cl
	
	cmp DWORD ARG2, 0
	jz .unset
.set:
	or DWORD [gaMM_Bitmap+edi*4], eax
	cmp DWORD [gaMM_Bitmap+edi*4], 0xFFFFFFFF
	jnz .ret
	
	; Set full flag in super bitmap
	mov ecx, edi
	shr edi, 5	; ECX /= 32
	and ecx, 31
	mov eax, 1
	shl eax, cl
	or DWORD [gaMM_SuperBitmap+edi*4], eax
	jmp .ret


.unset:
	not eax
	and DWORD [gaMM_Bitmap+edi], eax
	; Unset full flag in super bitmap
	mov ecx, edi
	shr edi, 5	; ECX /= 32
	and ecx, 31
	mov eax, 1
	shl eax, cl
	not eax
	and DWORD [gaMM_SuperBitmap+edi*4], eax
	

.ret:
	pop edi
	pop ebp
	ret

;; MM_INT_UpdateSuper
;; \brief Updates the super page bitmap for a page
MM_INT_UpdateSuper:
	push ebp
	mov ebp, esp
	push edi
	
	mov ecx, DWORD ARG1
	shr ecx, 12+5	; / 4096 / 32
	mov edi, ecx
	and ecx, 31
	shl edi, 2	; * 4
	mov eax, 1
	shl eax, cl
	
	cmp DWORD [gaMM_Bitmap+edi], 0xFFFFFFFF
	jz .full
.notfull:
	not	eax
	shr edi, 5	; / 32
	and DWORD [gaMM_SuperBitmap+edi], eax
	jmp	.ret
.full:
	shr edi, 5	; / 32
	or DWORD [gaMM_SuperBitmap+edi], eax

.ret:
	pop edi
	pop ebp
	ret
