; EmulOS 2
; By thePowersGang

; Thread Manager (named proc.asm for historical reasons)
; Threads:
; - Core (ThreadZero)
; - Machine
; - Video
; - Audio
; - Input
; - CPUs (Max `MAX_VCPUS`)

%include "common.inc.asm"
%include "proc.inc.asm"

%define MAX_CPUS	1
%define MAX_VCPUS	2
%define	NTHREADS	(1+4+MAX_VCPUS)
%define DEFAULT_QUANTUM	5

%define	MAGIC_SWITCH	0xFEED3EE1

[BITS 32]
[section .text]

[global Threads_Init]
Threads_Init:
	; Create thread 0 (core/control)
	mov WORD [gThreads_ThreadPool+tThread.State], 1
	mov WORD [gThreads_ThreadPool+tThread.Time], DEFAULT_QUANTUM
	; Apply to CPU0
	mov DWORD [gThreads_CPUs+0*tCPU_size+tCPU.NThreads], 1
	mov DWORD [gThreads_CPUs+0*tCPU_size+tCPU.Active], gThreads_ThreadPool
	mov DWORD [gThreads_CPUs+0*tCPU_size+tCPU.First], gThreads_ThreadPool
	; Start timers
	sti
	ret

;; Threads_Activate
;; \brief Intialises a thread
;; \param Number
;; \param Entry
;; \param NArgs
;; \param ...	Arguments
[global Threads_Activate]
Threads_Activate:
	FRAME_OPEN	0, 0, {esi, edi}

	mov esi, ARG1
	shl esi, tThread_bitsize
	add esi, gThreads_ThreadPool
	
	mov WORD [esi+tThread.State], 0
	mov WORD [esi+tThread.Time], DEFAULT_QUANTUM
	
	mov DWORD [esi+tThread.EIP], Threads_ActivateStub
	mov edi, ARG1
	shl edi, 12+1
	add edi, gThreads_Stacks+(1<<(12+1))
	
	mov ecx, ARG3	; NArgs
	lea eax, ARG4	; Arguments
	lea eax, [eax+ecx*4]
	test ecx, ecx
	jz .makestack.end
.makestack.loop:
	sub edi, 4
	sub eax, 4
	mov edx, DWORD [eax]
	mov DWORD [edi], edx
	dec ecx
	jnz .makestack.loop
.makestack.end:

	sub edi, 4
	mov eax, ARG2
	mov DWORD [edi], eax

	mov DWORD [esi+tThread.ESP], edi
	; TODO: Distribute threads throughout CPUs
	mov eax, 0
	mov WORD [esi+tThread.CPU], ax
	shl eax, 4	; EAX *= 16
	mov ecx, DWORD [gThreads_CPUs+eax+tCPU.First]
	mov DWORD [esi+tThread.Next], ecx
	mov DWORD [gThreads_CPUs+eax+tCPU.First], esi

	mov WORD [esi+tThread.State], 1

.ret:
	FRAME_CLOSE
	ret
;; Threads_ActivateStub
;; \brief Stub thread entrypoint that ACKs the interrupt and returns
Threads_ActivateStub:
	%if CPU0_USES_PIT
	mov dx, 0x20
	mov al, 0x20
	out dx, al
	%endif
	ret

; Increment the timestamp
; MUST be only called by CPU0
Threads_Timer:
	mov	eax, [gThreads_Timestamp]
	inc	eax
	mov	[gThreads_Timestamp], eax
	test eax, eax
	jnz .ret
	inc DWORD [gThreads_Timestamp+4]
.ret:
	ret

;; \brief Switch threads
;; \param EDI	CPU Number
;; 
;; Called on each scheduling timer tick to swap the current thread (if needed)
Threads_Switch:
	
	; If this is CPU0, call the timer
	test edi, edi
	jnz .skiptimer
	call Threads_Timer
.skiptimer:
	
	mov ebp, edi
	shl ebp, 4	; EBP *= 16
	add ebp, gThreads_CPUs	; EBP: CPU Structure
	;xchg bx, bx
	
	; EBP, EDI - Used
	
	; Check if the CPU has threads
	mov	ebx, DWORD [ebp+tCPU.NThreads]
	test ebx, ebx
	jz	.clihlt
	
	; Oh it does, now get the thread pointer
	mov	ebx, DWORD [ebp+tCPU.Active]
	; Check the remaining time steps
	mov ax, WORD [ebx+tThread.Time]
	test ax, ax
	jnz	.swapthread	; If there is no time remaining, go to the next thread
	; else, decrement and continue
	dec WORD [ebx+tThread.Time]
	
	ret
	
.swapthread:
	; EBX: Current thread
	; EBP: CPU
	
	; Check for thread switch
	call	GetEIP
	cmp eax, MAGIC_SWITCH
	jz	.ret	; If we just swapped, continue in the new task

	; Update thread state
	mov DWORD [ebx+tThread.EIP], eax
	mov DWORD [ebx+tThread.ESP], esp
	
	; Get next thread
	mov ebx, DWORD [ebx+tThread.Next]
	test ebx, ebx
	jnz	.updateCPU
	mov ebx, DWORD [ebp+tCPU.First]
.updateCPU:
	mov DWORD [ebp+tCPU.Active], ebx
	; Set new quantum
	mov WORD [ebx+tThread.Time], DEFAULT_QUANTUM
	
	; Return to task
	mov esp, DWORD [ebx+tThread.ESP]
	mov eax, MAGIC_SWITCH
	mov ecx, DWORD [ebx+tThread.EIP]
	jmp ecx
	; NEVER REACHED
	
.clihlt:
	cli
	hlt
.ret:
	ret

;; GetEIP
;; \brief Returns the address of the instruction folowing the call
GetEIP:
	mov eax, DWORD [esp]
	ret

[global TimerIRQCommon]
;; TimerIRQCommon
;; \brief Common section of the Timer IRQ
TimerIRQCommon:
	pusha	; 8 registers (a,b,c,d, di, si, bp, sp) = 8*4
	mov edi, [esp+8*4]	; Get CPU Number
	; Align ESP to 16-bytes
	mov eax, esp
.addpadding:
	push eax
	test esp, 15	; FXSAVE requires a 16-bit aligned reigon
	jnz	.addpadding
.pushfpu:
	sub esp, 512
	fxsave [esp]
	
	call	Threads_Switch
	
	; Reset PIT if needed
	%if CPU0_USES_PIT
	test edi, edi
	jnz .notCPU0
	mov dx, 0x20
	mov al, 0x20
	out dx, al
.notCPU0:
	%endif
	
	fxrstor [esp]
	add esp, 512
	pop eax	; Get ESP (undoes the aligning)
	mov esp, eax	; Update ESP from EAX
	
	popa
	add esp, 4	; Remove the CPU Number from the stack
	iret

%macro DEF_CPUTIMER	1
TimerIRQ_CPU%1:
	push %1
	jmp	TimerIRQCommon
%endmacro

%assign i	0
%rep MAX_CPUS
DEF_CPUTIMER	i
%assign i i+1
%endrep


[section .bss]
; CPUs - Contains the state for each CPU
gThreads_CPUs:
	resd	tCPU_size * MAX_CPUS
; Thread Pool - Statically Allocated threads
gThreads_ThreadPool:
	resb	NTHREADS * tThread_size
; Timestamp
gThreads_Timestamp:
	resd	1	; Lower 32 bits of timestamp
	resd	1	; Upper 32 bits
gThreads_Stacks:
	times NTHREADS	resb	0x2000
