; EmulOS 2
; By thePowersGang

; tui.asm
; - Text UI Control
%include "common.inc.asm"

[BITS 32]

%macro PUT_AL 0
	push edx
	mov dx, 0xE9
	out dx, al
	pop edx
%endmacro

[section .text]
;; TUI_ClearScreen
;; \brief Clears the screen
[global TUI_ClearScreen]
TUI_ClearScreen:
	push ebp
	mov ebp, esp
	push edi
	
	xor eax, eax
	mov ecx, 80*25/2
	mov edi, KERNEL_BASE + 0xB8000
	cld
	rep stosd
	
	pop edi
	pop ebp
	ret

;; TUI_Panic
;; \brief Displays an error message and then halts
[global TUI_Panic]
TUI_Panic:
	push ebp
	mov ebp, esp
	add esp, 16
	mov eax, [ebp+8]
	
	mov DWORD [esp], 0
	mov DWORD [esp+4], 24
	mov DWORD [esp+8], .str
	mov DWORD [esp+12], eax
	call TUI_Display
	
.hlt:
	hlt
	jmp .hlt
.str:	db	"--- PANIC: %s",0

;; TUI_Display
;; \brief Display text on the screenn
;; \param X
;; \param Y
;; \param Format
;; \param ...
[global TUI_Display]
TUI_Display:
	push ebp
	mov ebp, esp
	push esi
	push edi
	push ebx
	
	mov eax, [ebp+12]	; Get Y
	mov ebx, 80*2
	mul ebx	; EAX *= EBX
	mov edi, eax
	mov eax, [ebp+8]	; Get X
	lea edi, [edi+eax*2]	; Translate into Framebuffer Address
	add edi, KERNEL_BASE + 0xB8000
	mov esi, [ebp+16]	; Get Format String
	
	lea edx, [ebp+20]	; Arguments
	
	;xchg bx, bx
	
	; ESI - Input Format String
	; EDI - Framebuffer Address
	; EDX - First Argument's Address
	
.loop:
	lodsb
	test al, al
	jz	.ret
	
	cmp al, '%'
	jnz .literal
	;jmp .literal

	; Format Code
	lodsb
	test al, al
	jz	.ret
	
	; Hex
	cmp al, 'x'
	jnz .skipHex
	call TUI_INT_PutHex
.skipHex:
	cmp al, 's'
	jnz	.skipString
	call TUI_INT_PutString
.skipString:
	jmp .loop

.literal:
	mov BYTE [edi], al
	mov BYTE [edi+1], 0x07
	
	PUT_AL
	
	inc edi
	inc edi
	
	jmp .loop

.ret:
	mov al, 10
	PUT_AL
	pop ebx
	pop edi
	pop esi
	pop ebp
	ret

; EDI is Destination
; [EDX] is value
; EAX, ECX are free
TUI_INT_PutHex:
	mov ecx, DWORD [edx]
	add edx, 4
	; Nibble 1
	mov eax, ecx
	shr eax, 28
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi], al
	mov BYTE [edi+1], 0x07
	PUT_AL
	
	; Nibble 2
	mov eax, ecx
	shr eax, 24
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+2], al
	mov BYTE [edi+3], 0x07
	PUT_AL
	
	; Nibble 3
	mov eax, ecx
	shr eax, 20
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+4], al
	mov BYTE [edi+5], 0x07
	PUT_AL
	
	; Nibble 4
	mov eax, ecx
	shr eax, 16
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+6], al
	mov BYTE [edi+7], 0x07
	PUT_AL
	
	; Nibble 5
	mov eax, ecx
	shr eax, 12
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+8], al
	mov BYTE [edi+9], 0x07
	PUT_AL
	
	; Nibble 6
	mov eax, ecx
	shr eax, 8
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+10], al
	mov BYTE [edi+11], 0x07
	PUT_AL
	
	; Nibble 7
	mov eax, ecx
	shr eax, 4
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+12], al
	mov BYTE [edi+13], 0x07
	PUT_AL
	
	; Nibble 8
	mov eax, ecx
	and eax, 0xF
	mov al, BYTE [caTUI_HexDigits+eax]
	mov BYTE [edi+14], al
	mov BYTE [edi+15], 0x07
	PUT_AL
	
	add edi, 16
	ret


; EDI is Destination
; [EDX] is value
; EAX, ECX are free
TUI_INT_PutString:
	push esi	; Now ESI is also free
	mov esi, DWORD [edx]
	;xchg bx, bx
	add edx, 4
	test esi, esi
	jnz .loop
	mov esi, csTUI_NullString
.loop:
	lodsb
	test al, al
	jz	.ret
	mov BYTE [edi], al
	mov BYTE [edi+1], 0x07
	PUT_AL
	inc edi
	inc edi
	jmp .loop
.ret:
	pop esi
	ret
	
[section .rodata]
csTUI_NullString:
	db "(null)",0
caTUI_HexDigits:
	db "0123456789abcdef"
